import os
from pathlib import Path
from typing import Any, Callable, List, Optional, Tuple, Union

import numpy as np
import torch
from PIL import Image

from .utils import download_url
from .vision import VisionDataset


class PhotoTour(VisionDataset):
    """`Multi-view Stereo Correspondence <http://matthewalunbrown.com/patchdata/patchdata.html>`_ Dataset.

    .. note::

        We only provide the newer version of the dataset, since the authors state that it

            is more suitable for training descriptors based on difference of Gaussian, or Harris corners, as the
            patches are centred on real interest point detections, rather than being projections of 3D points as is the
            case in the old dataset.

        The original dataset is available under http://phototour.cs.washington.edu/patches/default.htm.


    Args:
        root (str or ``pathlib.Path``): Root directory where images are.
        name (string): Name of the dataset to load.
        transform (callable, optional): A function/transform that takes in a PIL image
            and returns a transformed version.
        download (bool, optional): If true, downloads the dataset from the internet and
            puts it in root directory. If dataset is already downloaded, it is not
            downloaded again.

    """

    urls = {
        "notredame_harris": [
            "http://matthewalunbrown.com/patchdata/notredame_harris.zip",
            "notredame_harris.zip",
            "69f8c90f78e171349abdf0307afefe4d",
        ],
        "yosemite_harris": [
            "http://matthewalunbrown.com/patchdata/yosemite_harris.zip",
            "yosemite_harris.zip",
            "a73253d1c6fbd3ba2613c45065c00d46",
        ],
        "liberty_harris": [
            "http://matthewalunbrown.com/patchdata/liberty_harris.zip",
            "liberty_harris.zip",
            "c731fcfb3abb4091110d0ae8c7ba182c",
        ],
        "notredame": [
            "http://icvl.ee.ic.ac.uk/vbalnt/notredame.zip",
            "notredame.zip",
            "509eda8535847b8c0a90bbb210c83484",
        ],
        "yosemite": ["http://icvl.ee.ic.ac.uk/vbalnt/yosemite.zip", "yosemite.zip", "533b2e8eb7ede31be40abc317b2fd4f0"],
        "liberty": ["http://icvl.ee.ic.ac.uk/vbalnt/liberty.zip", "liberty.zip", "fdd9152f138ea5ef2091746689176414"],
    }
    means = {
        "notredame": 0.4854,
        "yosemite": 0.4844,
        "liberty": 0.4437,
        "notredame_harris": 0.4854,
        "yosemite_harris": 0.4844,
        "liberty_harris": 0.4437,
    }
    stds = {
        "notredame": 0.1864,
        "yosemite": 0.1818,
        "liberty": 0.2019,
        "notredame_harris": 0.1864,
        "yosemite_harris": 0.1818,
        "liberty_harris": 0.2019,
    }
    lens = {
        "notredame": 468159,
        "yosemite": 633587,
        "liberty": 450092,
        "liberty_harris": 379587,
        "yosemite_harris": 450912,
        "notredame_harris": 325295,
    }
    image_ext = "bmp"
    info_file = "info.txt"
    matches_files = "m50_100000_100000_0.txt"

    def __init__(
        self,
        root: Union[str, Path],
        name: str,
        train: bool = True,
        transform: Optional[Callable] = None,
        download: bool = False,
    ) -> None:
        super().__init__(root, transform=transform)
        self.name = name
        self.data_dir = os.path.join(self.root, name)
        self.data_down = os.path.join(self.root, f"{name}.zip")
        self.data_file = os.path.join(self.root, f"{name}.pt")

        self.train = train
        self.mean = self.means[name]
        self.std = self.stds[name]

        if download:
            self.download()

        if not self._check_datafile_exists():
            self.cache()

        # load the serialized data
        self.data, self.labels, self.matches = torch.load(self.data_file, weights_only=True)

    def __getitem__(self, index: int) -> Union[torch.Tensor, Tuple[Any, Any, torch.Tensor]]:
        """
        Args:
            index (int): Index

        Returns:
            tuple: (data1, data2, matches)
        """
        if self.train:
            data = self.data[index]
            if self.transform is not None:
                data = self.transform(data)
            return data
        m = self.matches[index]
        data1, data2 = self.data[m[0]], self.data[m[1]]
        if self.transform is not None:
            data1 = self.transform(data1)
            data2 = self.transform(data2)
        return data1, data2, m[2]

    def __len__(self) -> int:
        return len(self.data if self.train else self.matches)

    def _check_datafile_exists(self) -> bool:
        return os.path.exists(self.data_file)

    def _check_downloaded(self) -> bool:
        return os.path.exists(self.data_dir)

    def download(self) -> None:
        if self._check_datafile_exists():
            print(f"# Found cached data {self.data_file}")
            return

        if not self._check_downloaded():
            # download files
            url = self.urls[self.name][0]
            filename = self.urls[self.name][1]
            md5 = self.urls[self.name][2]
            fpath = os.path.join(self.root, filename)

            download_url(url, self.root, filename, md5)

            print(f"# Extracting data {self.data_down}\n")

            import zipfile

            with zipfile.ZipFile(fpath, "r") as z:
                z.extractall(self.data_dir)

            os.unlink(fpath)

    def cache(self) -> None:
        # process and save as torch files
        print(f"# Caching data {self.data_file}")

        dataset = (
            read_image_file(self.data_dir, self.image_ext, self.lens[self.name]),
            read_info_file(self.data_dir, self.info_file),
            read_matches_files(self.data_dir, self.matches_files),
        )

        with open(self.data_file, "wb") as f:
            torch.save(dataset, f)

    def extra_repr(self) -> str:
        split = "Train" if self.train is True else "Test"
        return f"Split: {split}"


def read_image_file(data_dir: str, image_ext: str, n: int) -> torch.Tensor:
    """Return a Tensor containing the patches"""

    def PIL2array(_img: Image.Image) -> np.ndarray:
        """Convert PIL image type to numpy 2D array"""
        return np.array(_img.getdata(), dtype=np.uint8).reshape(64, 64)

    def find_files(_data_dir: str, _image_ext: str) -> List[str]:
        """Return a list with the file names of the images containing the patches"""
        files = []
        # find those files with the specified extension
        for file_dir in os.listdir(_data_dir):
            if file_dir.endswith(_image_ext):
                files.append(os.path.join(_data_dir, file_dir))
        return sorted(files)  # sort files in ascend order to keep relations

    patches = []
    list_files = find_files(data_dir, image_ext)

    for fpath in list_files:
        img = Image.open(fpath)
        for y in range(0, img.height, 64):
            for x in range(0, img.width, 64):
                patch = img.crop((x, y, x + 64, y + 64))
                patches.append(PIL2array(patch))
    return torch.ByteTensor(np.array(patches[:n]))


def read_info_file(data_dir: str, info_file: str) -> torch.Tensor:
    """Return a Tensor containing the list of labels
    Read the file and keep only the ID of the 3D point.
    """
    with open(os.path.join(data_dir, info_file)) as f:
        labels = [int(line.split()[0]) for line in f]
    return torch.LongTensor(labels)


def read_matches_files(data_dir: str, matches_file: str) -> torch.Tensor:
    """Return a Tensor containing the ground truth matches
    Read the file and keep only 3D point ID.
    Matches are represented with a 1, non matches with a 0.
    """
    matches = []
    with open(os.path.join(data_dir, matches_file)) as f:
        for line in f:
            line_split = line.split()
            matches.append([int(line_split[0]), int(line_split[3]), int(line_split[1] == line_split[4])])
    return torch.LongTensor(matches)



from pathlib import Path
from typing import Any, Callable, Optional, Tuple, Union

from PIL import Image

from .folder import find_classes, make_dataset
from .utils import download_and_extract_archive, verify_str_arg
from .vision import VisionDataset


class Imagenette(VisionDataset):
    """`Imagenette <https://github.com/fastai/imagenette#imagenette-1>`_ image classification dataset.

    Args:
        root (str or ``pathlib.Path``): Root directory of the Imagenette dataset.
        split (string, optional): The dataset split. Supports ``"train"`` (default), and ``"val"``.
        size (string, optional): The image size. Supports ``"full"`` (default), ``"320px"``, and ``"160px"``.
        download (bool, optional): If ``True``, downloads the dataset components and places them in ``root``. Already
            downloaded archives are not downloaded again.
        transform (callable, optional): A function/transform that takes in a PIL image and returns a transformed
            version, e.g. ``transforms.RandomCrop``.
        target_transform (callable, optional): A function/transform that takes in the target and transforms it.

     Attributes:
        classes (list): List of the class name tuples.
        class_to_idx (dict): Dict with items (class name, class index).
        wnids (list): List of the WordNet IDs.
        wnid_to_idx (dict): Dict with items (WordNet ID, class index).
    """

    _ARCHIVES = {
        "full": ("https://s3.amazonaws.com/fast-ai-imageclas/imagenette2.tgz", "fe2fc210e6bb7c5664d602c3cd71e612"),
        "320px": ("https://s3.amazonaws.com/fast-ai-imageclas/imagenette2-320.tgz", "3df6f0d01a2c9592104656642f5e78a3"),
        "160px": ("https://s3.amazonaws.com/fast-ai-imageclas/imagenette2-160.tgz", "e793b78cc4c9e9a4ccc0c1155377a412"),
    }
    _WNID_TO_CLASS = {
        "n01440764": ("tench", "Tinca tinca"),
        "n02102040": ("English springer", "English springer spaniel"),
        "n02979186": ("cassette player",),
        "n03000684": ("chain saw", "chainsaw"),
        "n03028079": ("church", "church building"),
        "n03394916": ("French horn", "horn"),
        "n03417042": ("garbage truck", "dustcart"),
        "n03425413": ("gas pump", "gasoline pump", "petrol pump", "island dispenser"),
        "n03445777": ("golf ball",),
        "n03888257": ("parachute", "chute"),
    }

    def __init__(
        self,
        root: Union[str, Path],
        split: str = "train",
        size: str = "full",
        download=False,
        transform: Optional[Callable] = None,
        target_transform: Optional[Callable] = None,
    ) -> None:
        super().__init__(root, transform=transform, target_transform=target_transform)

        self._split = verify_str_arg(split, "split", ["train", "val"])
        self._size = verify_str_arg(size, "size", ["full", "320px", "160px"])

        self._url, self._md5 = self._ARCHIVES[self._size]
        self._size_root = Path(self.root) / Path(self._url).stem
        self._image_root = str(self._size_root / self._split)

        if download:
            self._download()
        elif not self._check_exists():
            raise RuntimeError("Dataset not found. You can use download=True to download it.")

        self.wnids, self.wnid_to_idx = find_classes(self._image_root)
        self.classes = [self._WNID_TO_CLASS[wnid] for wnid in self.wnids]
        self.class_to_idx = {
            class_name: idx for wnid, idx in self.wnid_to_idx.items() for class_name in self._WNID_TO_CLASS[wnid]
        }
        self._samples = make_dataset(self._image_root, self.wnid_to_idx, extensions=".jpeg")

    def _check_exists(self) -> bool:
        return self._size_root.exists()

    def _download(self):
        if self._check_exists():
            raise RuntimeError(
                f"The directory {self._size_root} already exists. "
                f"If you want to re-download or re-extract the images, delete the directory."
            )

        download_and_extract_archive(self._url, self.root, md5=self._md5)

    def __getitem__(self, idx: int) -> Tuple[Any, Any]:
        path, label = self._samples[idx]
        image = Image.open(path).convert("RGB")

        if self.transform is not None:
            image = self.transform(image)

        if self.target_transform is not None:
            label = self.target_transform(label)

        return image, label

    def __len__(self) -> int:
        return len(self._samples)



import os
from os.path import abspath, expanduser
from pathlib import Path

from typing import Any, Callable, Dict, List, Optional, Tuple, Union

import torch
from PIL import Image

from .utils import download_and_extract_archive, download_file_from_google_drive, extract_archive, verify_str_arg
from .vision import VisionDataset


class WIDERFace(VisionDataset):
    """`WIDERFace <http://shuoyang1213.me/WIDERFACE/>`_ Dataset.

    Args:
        root (str or ``pathlib.Path``): Root directory where images and annotations are downloaded to.
            Expects the following folder structure if download=False:

            .. code::

                <root>
                    └── widerface
                        ├── wider_face_split ('wider_face_split.zip' if compressed)
                        ├── WIDER_train ('WIDER_train.zip' if compressed)
                        ├── WIDER_val ('WIDER_val.zip' if compressed)
                        └── WIDER_test ('WIDER_test.zip' if compressed)
        split (string): The dataset split to use. One of {``train``, ``val``, ``test``}.
            Defaults to ``train``.
        transform (callable, optional): A function/transform that takes in a PIL image
            and returns a transformed version. E.g, ``transforms.RandomCrop``
        target_transform (callable, optional): A function/transform that takes in the
            target and transforms it.
        download (bool, optional): If true, downloads the dataset from the internet and
            puts it in root directory. If dataset is already downloaded, it is not
            downloaded again.

            .. warning::

                To download the dataset `gdown <https://github.com/wkentaro/gdown>`_ is required.

    """

    BASE_FOLDER = "widerface"
    FILE_LIST = [
        # File ID                             MD5 Hash                            Filename
        ("15hGDLhsx8bLgLcIRD5DhYt5iBxnjNF1M", "3fedf70df600953d25982bcd13d91ba2", "WIDER_train.zip"),
        ("1GUCogbp16PMGa39thoMMeWxp7Rp5oM8Q", "dfa7d7e790efa35df3788964cf0bbaea", "WIDER_val.zip"),
        ("1HIfDbVEWKmsYKJZm4lchTBDLW5N7dY5T", "e5d8f4248ed24c334bbd12f49c29dd40", "WIDER_test.zip"),
    ]
    ANNOTATIONS_FILE = (
        "http://shuoyang1213.me/WIDERFACE/support/bbx_annotation/wider_face_split.zip",
        "0e3767bcf0e326556d407bf5bff5d27c",
        "wider_face_split.zip",
    )

    def __init__(
        self,
        root: Union[str, Path],
        split: str = "train",
        transform: Optional[Callable] = None,
        target_transform: Optional[Callable] = None,
        download: bool = False,
    ) -> None:
        super().__init__(
            root=os.path.join(root, self.BASE_FOLDER), transform=transform, target_transform=target_transform
        )
        # check arguments
        self.split = verify_str_arg(split, "split", ("train", "val", "test"))

        if download:
            self.download()

        if not self._check_integrity():
            raise RuntimeError("Dataset not found or corrupted. You can use download=True to download and prepare it")

        self.img_info: List[Dict[str, Union[str, Dict[str, torch.Tensor]]]] = []
        if self.split in ("train", "val"):
            self.parse_train_val_annotations_file()
        else:
            self.parse_test_annotations_file()

    def __getitem__(self, index: int) -> Tuple[Any, Any]:
        """
        Args:
            index (int): Index

        Returns:
            tuple: (image, target) where target is a dict of annotations for all faces in the image.
            target=None for the test split.
        """

        # stay consistent with other datasets and return a PIL Image
        img = Image.open(self.img_info[index]["img_path"])

        if self.transform is not None:
            img = self.transform(img)

        target = None if self.split == "test" else self.img_info[index]["annotations"]
        if self.target_transform is not None:
            target = self.target_transform(target)

        return img, target

    def __len__(self) -> int:
        return len(self.img_info)

    def extra_repr(self) -> str:
        lines = ["Split: {split}"]
        return "\n".join(lines).format(**self.__dict__)

    def parse_train_val_annotations_file(self) -> None:
        filename = "wider_face_train_bbx_gt.txt" if self.split == "train" else "wider_face_val_bbx_gt.txt"
        filepath = os.path.join(self.root, "wider_face_split", filename)

        with open(filepath) as f:
            lines = f.readlines()
            file_name_line, num_boxes_line, box_annotation_line = True, False, False
            num_boxes, box_counter = 0, 0
            labels = []
            for line in lines:
                line = line.rstrip()
                if file_name_line:
                    img_path = os.path.join(self.root, "WIDER_" + self.split, "images", line)
                    img_path = abspath(expanduser(img_path))
                    file_name_line = False
                    num_boxes_line = True
                elif num_boxes_line:
                    num_boxes = int(line)
                    num_boxes_line = False
                    box_annotation_line = True
                elif box_annotation_line:
                    box_counter += 1
                    line_split = line.split(" ")
                    line_values = [int(x) for x in line_split]
                    labels.append(line_values)
                    if box_counter >= num_boxes:
                        box_annotation_line = False
                        file_name_line = True
                        labels_tensor = torch.tensor(labels)
                        self.img_info.append(
                            {
                                "img_path": img_path,
                                "annotations": {
                                    "bbox": labels_tensor[:, 0:4].clone(),  # x, y, width, height
                                    "blur": labels_tensor[:, 4].clone(),
                                    "expression": labels_tensor[:, 5].clone(),
                                    "illumination": labels_tensor[:, 6].clone(),
                                    "occlusion": labels_tensor[:, 7].clone(),
                                    "pose": labels_tensor[:, 8].clone(),
                                    "invalid": labels_tensor[:, 9].clone(),
                                },
                            }
                        )
                        box_counter = 0
                        labels.clear()
                else:
                    raise RuntimeError(f"Error parsing annotation file {filepath}")

    def parse_test_annotations_file(self) -> None:
        filepath = os.path.join(self.root, "wider_face_split", "wider_face_test_filelist.txt")
        filepath = abspath(expanduser(filepath))
        with open(filepath) as f:
            lines = f.readlines()
            for line in lines:
                line = line.rstrip()
                img_path = os.path.join(self.root, "WIDER_test", "images", line)
                img_path = abspath(expanduser(img_path))
                self.img_info.append({"img_path": img_path})

    def _check_integrity(self) -> bool:
        # Allow original archive to be deleted (zip). Only need the extracted images
        all_files = self.FILE_LIST.copy()
        all_files.append(self.ANNOTATIONS_FILE)
        for (_, md5, filename) in all_files:
            file, ext = os.path.splitext(filename)
            extracted_dir = os.path.join(self.root, file)
            if not os.path.exists(extracted_dir):
                return False
        return True

    def download(self) -> None:
        if self._check_integrity():
            print("Files already downloaded and verified")
            return

        # download and extract image data
        for (file_id, md5, filename) in self.FILE_LIST:
            download_file_from_google_drive(file_id, self.root, filename, md5)
            filepath = os.path.join(self.root, filename)
            extract_archive(filepath)

        # download and extract annotation files
        download_and_extract_archive(
            url=self.ANNOTATIONS_FILE[0], download_root=self.root, md5=self.ANNOTATIONS_FILE[1]
        )



import unittest

import pytest
import torch

from torchvision.models.maxvit import SwapAxes, WindowDepartition, WindowPartition


class MaxvitTester(unittest.TestCase):
    def test_maxvit_window_partition(self):
        input_shape = (1, 3, 224, 224)
        partition_size = 7
        n_partitions = input_shape[3] // partition_size

        x = torch.randn(input_shape)

        partition = WindowPartition()
        departition = WindowDepartition()

        x_hat = partition(x, partition_size)
        x_hat = departition(x_hat, partition_size, n_partitions, n_partitions)

        torch.testing.assert_close(x, x_hat)

    def test_maxvit_grid_partition(self):
        input_shape = (1, 3, 224, 224)
        partition_size = 7
        n_partitions = input_shape[3] // partition_size

        x = torch.randn(input_shape)
        pre_swap = SwapAxes(-2, -3)
        post_swap = SwapAxes(-2, -3)

        partition = WindowPartition()
        departition = WindowDepartition()

        x_hat = partition(x, n_partitions)
        x_hat = pre_swap(x_hat)
        x_hat = post_swap(x_hat)
        x_hat = departition(x_hat, n_partitions, partition_size, partition_size)

        torch.testing.assert_close(x, x_hat)


if __name__ == "__main__":
    pytest.main([__file__])



import functools
from typing import Any, Callable, Dict, List, Optional, Sequence, Type, Union

import torch
from torchvision import tv_tensors

_FillType = Union[int, float, Sequence[int], Sequence[float], None]
_FillTypeJIT = Optional[List[float]]


def is_pure_tensor(inpt: Any) -> bool:
    return isinstance(inpt, torch.Tensor) and not isinstance(inpt, tv_tensors.TVTensor)


# {functional: {input_type: type_specific_kernel}}
_KERNEL_REGISTRY: Dict[Callable, Dict[Type, Callable]] = {}


def _kernel_tv_tensor_wrapper(kernel):
    @functools.wraps(kernel)
    def wrapper(inpt, *args, **kwargs):
        # If you're wondering whether we could / should get rid of this wrapper,
        # the answer is no: we want to pass pure Tensors to avoid the overhead
        # of the __torch_function__ machinery. Note that this is always valid,
        # regardless of whether we override __torch_function__ in our base class
        # or not.
        # Also, even if we didn't call `as_subclass` here, we would still need
        # this wrapper to call wrap(), because the TVTensor type would be
        # lost after the first operation due to our own __torch_function__
        # logic.
        output = kernel(inpt.as_subclass(torch.Tensor), *args, **kwargs)
        return tv_tensors.wrap(output, like=inpt)

    return wrapper


def _register_kernel_internal(functional, input_type, *, tv_tensor_wrapper=True):
    registry = _KERNEL_REGISTRY.setdefault(functional, {})
    if input_type in registry:
        raise ValueError(f"Functional {functional} already has a kernel registered for type {input_type}.")

    def decorator(kernel):
        registry[input_type] = (
            _kernel_tv_tensor_wrapper(kernel)
            if issubclass(input_type, tv_tensors.TVTensor) and tv_tensor_wrapper
            else kernel
        )
        return kernel

    return decorator


def _name_to_functional(name):
    import torchvision.transforms.v2.functional  # noqa

    try:
        return getattr(torchvision.transforms.v2.functional, name)
    except AttributeError:
        raise ValueError(
            f"Could not find functional with name '{name}' in torchvision.transforms.v2.functional."
        ) from None


_BUILTIN_DATAPOINT_TYPES = {
    obj for obj in tv_tensors.__dict__.values() if isinstance(obj, type) and issubclass(obj, tv_tensors.TVTensor)
}


def register_kernel(functional, tv_tensor_cls):
    """Decorate a kernel to register it for a functional and a (custom) tv_tensor type.

    See :ref:`sphx_glr_auto_examples_transforms_plot_custom_tv_tensors.py` for usage
    details.
    """
    if isinstance(functional, str):
        functional = _name_to_functional(name=functional)
    elif not (
        callable(functional)
        and getattr(functional, "__module__", "").startswith("torchvision.transforms.v2.functional")
    ):
        raise ValueError(
            f"Kernels can only be registered on functionals from the torchvision.transforms.v2.functional namespace, "
            f"but got {functional}."
        )

    if not (isinstance(tv_tensor_cls, type) and issubclass(tv_tensor_cls, tv_tensors.TVTensor)):
        raise ValueError(
            f"Kernels can only be registered for subclasses of torchvision.tv_tensors.TVTensor, "
            f"but got {tv_tensor_cls}."
        )

    if tv_tensor_cls in _BUILTIN_DATAPOINT_TYPES:
        raise ValueError(f"Kernels cannot be registered for the builtin tv_tensor classes, but got {tv_tensor_cls}")

    return _register_kernel_internal(functional, tv_tensor_cls, tv_tensor_wrapper=False)


def _get_kernel(functional, input_type, *, allow_passthrough=False):
    registry = _KERNEL_REGISTRY.get(functional)
    if not registry:
        raise ValueError(f"No kernel registered for functional {functional.__name__}.")

    for cls in input_type.__mro__:
        if cls in registry:
            return registry[cls]
        elif cls is tv_tensors.TVTensor:
            # We don't want user-defined tv_tensors to dispatch to the pure Tensor kernels, so we explicit stop the
            # MRO traversal before hitting torch.Tensor. We can even stop at tv_tensors.TVTensor, since we don't
            # allow kernels to be registered for tv_tensors.TVTensor anyway.
            break

    if allow_passthrough:
        return lambda inpt, *args, **kwargs: inpt

    raise TypeError(
        f"Functional F.{functional.__name__} supports inputs of type {registry.keys()}, "
        f"but got {input_type} instead."
    )


# This basically replicates _register_kernel_internal, but with a specialized wrapper for five_crop / ten_crop
# We could get rid of this by letting _register_kernel_internal take arbitrary functionals rather than wrap_kernel: bool
def _register_five_ten_crop_kernel_internal(functional, input_type):
    registry = _KERNEL_REGISTRY.setdefault(functional, {})
    if input_type in registry:
        raise TypeError(f"Functional '{functional}' already has a kernel registered for type '{input_type}'.")

    def wrap(kernel):
        @functools.wraps(kernel)
        def wrapper(inpt, *args, **kwargs):
            output = kernel(inpt, *args, **kwargs)
            container_type = type(output)
            return container_type(tv_tensors.wrap(o, like=inpt) for o in output)

        return wrapper

    def decorator(kernel):
        registry[input_type] = wrap(kernel) if issubclass(input_type, tv_tensors.TVTensor) else kernel
        return kernel

    return decorator



import math
import numbers
import warnings
from typing import Any, Callable, Dict, List, Sequence, Tuple, Union

import PIL.Image
import torch
from torch.nn.functional import one_hot
from torch.utils._pytree import tree_flatten, tree_unflatten
from torchvision import transforms as _transforms, tv_tensors
from torchvision.transforms.v2 import functional as F

from ._transform import _RandomApplyTransform, Transform
from ._utils import _check_sequence_input, _parse_labels_getter, has_any, is_pure_tensor, query_chw, query_size


class RandomErasing(_RandomApplyTransform):
    """Randomly select a rectangle region in the input image or video and erase its pixels.

    This transform does not support PIL Image.
    'Random Erasing Data Augmentation' by Zhong et al. See https://arxiv.org/abs/1708.04896

    Args:
        p (float, optional): probability that the random erasing operation will be performed.
        scale (tuple of float, optional): range of proportion of erased area against input image.
        ratio (tuple of float, optional): range of aspect ratio of erased area.
        value (number or tuple of numbers): erasing value. Default is 0. If a single int, it is used to
            erase all pixels. If a tuple of length 3, it is used to erase
            R, G, B channels respectively.
            If a str of 'random', erasing each pixel with random values.
        inplace (bool, optional): boolean to make this transform inplace. Default set to False.

    Returns:
        Erased input.

    Example:
        >>> from torchvision.transforms import v2 as transforms
        >>>
        >>> transform = transforms.Compose([
        >>>   transforms.RandomHorizontalFlip(),
        >>>   transforms.PILToTensor(),
        >>>   transforms.ConvertImageDtype(torch.float),
        >>>   transforms.Normalize((0.485, 0.456, 0.406), (0.229, 0.224, 0.225)),
        >>>   transforms.RandomErasing(),
        >>> ])
    """

    _v1_transform_cls = _transforms.RandomErasing

    def _extract_params_for_v1_transform(self) -> Dict[str, Any]:
        return dict(
            super()._extract_params_for_v1_transform(),
            value="random" if self.value is None else self.value,
        )

    def __init__(
        self,
        p: float = 0.5,
        scale: Tuple[float, float] = (0.02, 0.33),
        ratio: Tuple[float, float] = (0.3, 3.3),
        value: float = 0.0,
        inplace: bool = False,
    ):
        super().__init__(p=p)
        if not isinstance(value, (numbers.Number, str, tuple, list)):
            raise TypeError("Argument value should be either a number or str or a sequence")
        if isinstance(value, str) and value != "random":
            raise ValueError("If value is str, it should be 'random'")
        if not isinstance(scale, (tuple, list)):
            raise TypeError("Scale should be a sequence")
        if not isinstance(ratio, (tuple, list)):
            raise TypeError("Ratio should be a sequence")
        if (scale[0] > scale[1]) or (ratio[0] > ratio[1]):
            warnings.warn("Scale and ratio should be of kind (min, max)")
        if scale[0] < 0 or scale[1] > 1:
            raise ValueError("Scale should be between 0 and 1")
        self.scale = scale
        self.ratio = ratio
        if isinstance(value, (int, float)):
            self.value = [float(value)]
        elif isinstance(value, str):
            self.value = None
        elif isinstance(value, (list, tuple)):
            self.value = [float(v) for v in value]
        else:
            self.value = value
        self.inplace = inplace

        self._log_ratio = torch.log(torch.tensor(self.ratio))

    def _call_kernel(self, functional: Callable, inpt: Any, *args: Any, **kwargs: Any) -> Any:
        if isinstance(inpt, (tv_tensors.BoundingBoxes, tv_tensors.Mask)):
            warnings.warn(
                f"{type(self).__name__}() is currently passing through inputs of type "
                f"tv_tensors.{type(inpt).__name__}. This will likely change in the future."
            )
        return super()._call_kernel(functional, inpt, *args, **kwargs)

    def _get_params(self, flat_inputs: List[Any]) -> Dict[str, Any]:
        img_c, img_h, img_w = query_chw(flat_inputs)

        if self.value is not None and not (len(self.value) in (1, img_c)):
            raise ValueError(
                f"If value is a sequence, it should have either a single value or {img_c} (number of inpt channels)"
            )

        area = img_h * img_w

        log_ratio = self._log_ratio
        for _ in range(10):
            erase_area = area * torch.empty(1).uniform_(self.scale[0], self.scale[1]).item()
            aspect_ratio = torch.exp(
                torch.empty(1).uniform_(
                    log_ratio[0],  # type: ignore[arg-type]
                    log_ratio[1],  # type: ignore[arg-type]
                )
            ).item()

            h = int(round(math.sqrt(erase_area * aspect_ratio)))
            w = int(round(math.sqrt(erase_area / aspect_ratio)))
            if not (h < img_h and w < img_w):
                continue

            if self.value is None:
                v = torch.empty([img_c, h, w], dtype=torch.float32).normal_()
            else:
                v = torch.tensor(self.value)[:, None, None]

            i = torch.randint(0, img_h - h + 1, size=(1,)).item()
            j = torch.randint(0, img_w - w + 1, size=(1,)).item()
            break
        else:
            i, j, h, w, v = 0, 0, img_h, img_w, None

        return dict(i=i, j=j, h=h, w=w, v=v)

    def _transform(self, inpt: Any, params: Dict[str, Any]) -> Any:
        if params["v"] is not None:
            inpt = self._call_kernel(F.erase, inpt, **params, inplace=self.inplace)

        return inpt


class _BaseMixUpCutMix(Transform):
    def __init__(self, *, alpha: float = 1.0, num_classes: int, labels_getter="default") -> None:
        super().__init__()
        self.alpha = float(alpha)
        self._dist = torch.distributions.Beta(torch.tensor([alpha]), torch.tensor([alpha]))

        self.num_classes = num_classes

        self._labels_getter = _parse_labels_getter(labels_getter)

    def forward(self, *inputs):
        inputs = inputs if len(inputs) > 1 else inputs[0]
        flat_inputs, spec = tree_flatten(inputs)
        needs_transform_list = self._needs_transform_list(flat_inputs)

        if has_any(flat_inputs, PIL.Image.Image, tv_tensors.BoundingBoxes, tv_tensors.Mask):
            raise ValueError(f"{type(self).__name__}() does not support PIL images, bounding boxes and masks.")

        labels = self._labels_getter(inputs)
        if not isinstance(labels, torch.Tensor):
            raise ValueError(f"The labels must be a tensor, but got {type(labels)} instead.")
        elif labels.ndim != 1:
            raise ValueError(
                f"labels tensor should be of shape (batch_size,) " f"but got shape {labels.shape} instead."
            )

        params = {
            "labels": labels,
            "batch_size": labels.shape[0],
            **self._get_params(
                [inpt for (inpt, needs_transform) in zip(flat_inputs, needs_transform_list) if needs_transform]
            ),
        }

        # By default, the labels will be False inside needs_transform_list, since they are a torch.Tensor coming
        # after an image or video. However, we need to handle them in _transform, so we make sure to set them to True
        needs_transform_list[next(idx for idx, inpt in enumerate(flat_inputs) if inpt is labels)] = True
        flat_outputs = [
            self._transform(inpt, params) if needs_transform else inpt
            for (inpt, needs_transform) in zip(flat_inputs, needs_transform_list)
        ]

        return tree_unflatten(flat_outputs, spec)

    def _check_image_or_video(self, inpt: torch.Tensor, *, batch_size: int):
        expected_num_dims = 5 if isinstance(inpt, tv_tensors.Video) else 4
        if inpt.ndim != expected_num_dims:
            raise ValueError(
                f"Expected a batched input with {expected_num_dims} dims, but got {inpt.ndim} dimensions instead."
            )
        if inpt.shape[0] != batch_size:
            raise ValueError(
                f"The batch size of the image or video does not match the batch size of the labels: "
                f"{inpt.shape[0]} != {batch_size}."
            )

    def _mixup_label(self, label: torch.Tensor, *, lam: float) -> torch.Tensor:
        label = one_hot(label, num_classes=self.num_classes)
        if not label.dtype.is_floating_point:
            label = label.float()
        return label.roll(1, 0).mul_(1.0 - lam).add_(label.mul(lam))


class MixUp(_BaseMixUpCutMix):
    """Apply MixUp to the provided batch of images and labels.

    Paper: `mixup: Beyond Empirical Risk Minimization <https://arxiv.org/abs/1710.09412>`_.

    .. note::
        This transform is meant to be used on **batches** of samples, not
        individual images. See
        :ref:`sphx_glr_auto_examples_transforms_plot_cutmix_mixup.py` for detailed usage
        examples.
        The sample pairing is deterministic and done by matching consecutive
        samples in the batch, so the batch needs to be shuffled (this is an
        implementation detail, not a guaranteed convention.)

    In the input, the labels are expected to be a tensor of shape ``(batch_size,)``. They will be transformed
    into a tensor of shape ``(batch_size, num_classes)``.

    Args:
        alpha (float, optional): hyperparameter of the Beta distribution used for mixup. Default is 1.
        num_classes (int): number of classes in the batch. Used for one-hot-encoding.
        labels_getter (callable or "default", optional): indicates how to identify the labels in the input.
            By default, this will pick the second parameter as the labels if it's a tensor. This covers the most
            common scenario where this transform is called as ``MixUp()(imgs_batch, labels_batch)``.
            It can also be a callable that takes the same input as the transform, and returns the labels.
    """

    def _get_params(self, flat_inputs: List[Any]) -> Dict[str, Any]:
        return dict(lam=float(self._dist.sample(())))  # type: ignore[arg-type]

    def _transform(self, inpt: Any, params: Dict[str, Any]) -> Any:
        lam = params["lam"]

        if inpt is params["labels"]:
            return self._mixup_label(inpt, lam=lam)
        elif isinstance(inpt, (tv_tensors.Image, tv_tensors.Video)) or is_pure_tensor(inpt):
            self._check_image_or_video(inpt, batch_size=params["batch_size"])

            output = inpt.roll(1, 0).mul_(1.0 - lam).add_(inpt.mul(lam))

            if isinstance(inpt, (tv_tensors.Image, tv_tensors.Video)):
                output = tv_tensors.wrap(output, like=inpt)

            return output
        else:
            return inpt


class CutMix(_BaseMixUpCutMix):
    """Apply CutMix to the provided batch of images and labels.

    Paper: `CutMix: Regularization Strategy to Train Strong Classifiers with Localizable Features
    <https://arxiv.org/abs/1905.04899>`_.

    .. note::
        This transform is meant to be used on **batches** of samples, not
        individual images. See
        :ref:`sphx_glr_auto_examples_transforms_plot_cutmix_mixup.py` for detailed usage
        examples.
        The sample pairing is deterministic and done by matching consecutive
        samples in the batch, so the batch needs to be shuffled (this is an
        implementation detail, not a guaranteed convention.)

    In the input, the labels are expected to be a tensor of shape ``(batch_size,)``. They will be transformed
    into a tensor of shape ``(batch_size, num_classes)``.

    Args:
        alpha (float, optional): hyperparameter of the Beta distribution used for mixup. Default is 1.
        num_classes (int): number of classes in the batch. Used for one-hot-encoding.
        labels_getter (callable or "default", optional): indicates how to identify the labels in the input.
            By default, this will pick the second parameter as the labels if it's a tensor. This covers the most
            common scenario where this transform is called as ``CutMix()(imgs_batch, labels_batch)``.
            It can also be a callable that takes the same input as the transform, and returns the labels.
    """

    def _get_params(self, flat_inputs: List[Any]) -> Dict[str, Any]:
        lam = float(self._dist.sample(()))  # type: ignore[arg-type]

        H, W = query_size(flat_inputs)

        r_x = torch.randint(W, size=(1,))
        r_y = torch.randint(H, size=(1,))

        r = 0.5 * math.sqrt(1.0 - lam)
        r_w_half = int(r * W)
        r_h_half = int(r * H)

        x1 = int(torch.clamp(r_x - r_w_half, min=0))
        y1 = int(torch.clamp(r_y - r_h_half, min=0))
        x2 = int(torch.clamp(r_x + r_w_half, max=W))
        y2 = int(torch.clamp(r_y + r_h_half, max=H))
        box = (x1, y1, x2, y2)

        lam_adjusted = float(1.0 - (x2 - x1) * (y2 - y1) / (W * H))

        return dict(box=box, lam_adjusted=lam_adjusted)

    def _transform(self, inpt: Any, params: Dict[str, Any]) -> Any:
        if inpt is params["labels"]:
            return self._mixup_label(inpt, lam=params["lam_adjusted"])
        elif isinstance(inpt, (tv_tensors.Image, tv_tensors.Video)) or is_pure_tensor(inpt):
            self._check_image_or_video(inpt, batch_size=params["batch_size"])

            x1, y1, x2, y2 = params["box"]
            rolled = inpt.roll(1, 0)
            output = inpt.clone()
            output[..., y1:y2, x1:x2] = rolled[..., y1:y2, x1:x2]

            if isinstance(inpt, (tv_tensors.Image, tv_tensors.Video)):
                output = tv_tensors.wrap(output, like=inpt)

            return output
        else:
            return inpt


class JPEG(Transform):
    """Apply JPEG compression and decompression to the given images.

    If the input is a :class:`torch.Tensor`, it is expected
    to be of dtype uint8, on CPU, and have [..., 3 or 1, H, W] shape,
    where ... means an arbitrary number of leading dimensions.

    Args:
        quality (sequence or number): JPEG quality, from 1 to 100. Lower means more compression.
            If quality is a sequence like (min, max), it specifies the range of JPEG quality to
            randomly select from (inclusive of both ends).

    Returns:
        image with JPEG compression.
    """

    def __init__(self, quality: Union[int, Sequence[int]]):
        super().__init__()
        if isinstance(quality, int):
            quality = [quality, quality]
        else:
            _check_sequence_input(quality, "quality", req_sizes=(2,))

        if not (1 <= quality[0] <= quality[1] <= 100 and isinstance(quality[0], int) and isinstance(quality[1], int)):
            raise ValueError(f"quality must be an integer from 1 to 100, got {quality =}")

        self.quality = quality

    def _get_params(self, flat_inputs: List[Any]) -> Dict[str, Any]:
        quality = torch.randint(self.quality[0], self.quality[1] + 1, ()).item()
        return dict(quality=quality)

    def _transform(self, inpt: Any, params: Dict[str, Any]) -> Any:
        return self._call_kernel(F.jpeg, inpt, quality=params["quality"])



from functools import partial
from typing import Any, Optional

from torch import nn

from ...transforms._presets import SemanticSegmentation
from .._api import register_model, Weights, WeightsEnum
from .._meta import _VOC_CATEGORIES
from .._utils import _ovewrite_value_param, handle_legacy_interface, IntermediateLayerGetter
from ..resnet import ResNet, resnet101, ResNet101_Weights, resnet50, ResNet50_Weights
from ._utils import _SimpleSegmentationModel


__all__ = ["FCN", "FCN_ResNet50_Weights", "FCN_ResNet101_Weights", "fcn_resnet50", "fcn_resnet101"]


class FCN(_SimpleSegmentationModel):
    """
    Implements FCN model from
    `"Fully Convolutional Networks for Semantic Segmentation"
    <https://arxiv.org/abs/1411.4038>`_.

    Args:
        backbone (nn.Module): the network used to compute the features for the model.
            The backbone should return an OrderedDict[Tensor], with the key being
            "out" for the last feature map used, and "aux" if an auxiliary classifier
            is used.
        classifier (nn.Module): module that takes the "out" element returned from
            the backbone and returns a dense prediction.
        aux_classifier (nn.Module, optional): auxiliary classifier used during training
    """

    pass


class FCNHead(nn.Sequential):
    def __init__(self, in_channels: int, channels: int) -> None:
        inter_channels = in_channels // 4
        layers = [
            nn.Conv2d(in_channels, inter_channels, 3, padding=1, bias=False),
            nn.BatchNorm2d(inter_channels),
            nn.ReLU(),
            nn.Dropout(0.1),
            nn.Conv2d(inter_channels, channels, 1),
        ]

        super().__init__(*layers)


_COMMON_META = {
    "categories": _VOC_CATEGORIES,
    "min_size": (1, 1),
    "_docs": """
        These weights were trained on a subset of COCO, using only the 20 categories that are present in the Pascal VOC
        dataset.
    """,
}


class FCN_ResNet50_Weights(WeightsEnum):
    COCO_WITH_VOC_LABELS_V1 = Weights(
        url="https://download.pytorch.org/models/fcn_resnet50_coco-1167a1af.pth",
        transforms=partial(SemanticSegmentation, resize_size=520),
        meta={
            **_COMMON_META,
            "num_params": 35322218,
            "recipe": "https://github.com/pytorch/vision/tree/main/references/segmentation#fcn_resnet50",
            "_metrics": {
                "COCO-val2017-VOC-labels": {
                    "miou": 60.5,
                    "pixel_acc": 91.4,
                }
            },
            "_ops": 152.717,
            "_file_size": 135.009,
        },
    )
    DEFAULT = COCO_WITH_VOC_LABELS_V1


class FCN_ResNet101_Weights(WeightsEnum):
    COCO_WITH_VOC_LABELS_V1 = Weights(
        url="https://download.pytorch.org/models/fcn_resnet101_coco-7ecb50ca.pth",
        transforms=partial(SemanticSegmentation, resize_size=520),
        meta={
            **_COMMON_META,
            "num_params": 54314346,
            "recipe": "https://github.com/pytorch/vision/tree/main/references/segmentation#deeplabv3_resnet101",
            "_metrics": {
                "COCO-val2017-VOC-labels": {
                    "miou": 63.7,
                    "pixel_acc": 91.9,
                }
            },
            "_ops": 232.738,
            "_file_size": 207.711,
        },
    )
    DEFAULT = COCO_WITH_VOC_LABELS_V1


def _fcn_resnet(
    backbone: ResNet,
    num_classes: int,
    aux: Optional[bool],
) -> FCN:
    return_layers = {"layer4": "out"}
    if aux:
        return_layers["layer3"] = "aux"
    backbone = IntermediateLayerGetter(backbone, return_layers=return_layers)

    aux_classifier = FCNHead(1024, num_classes) if aux else None
    classifier = FCNHead(2048, num_classes)
    return FCN(backbone, classifier, aux_classifier)


@register_model()
@handle_legacy_interface(
    weights=("pretrained", FCN_ResNet50_Weights.COCO_WITH_VOC_LABELS_V1),
    weights_backbone=("pretrained_backbone", ResNet50_Weights.IMAGENET1K_V1),
)
def fcn_resnet50(
    *,
    weights: Optional[FCN_ResNet50_Weights] = None,
    progress: bool = True,
    num_classes: Optional[int] = None,
    aux_loss: Optional[bool] = None,
    weights_backbone: Optional[ResNet50_Weights] = ResNet50_Weights.IMAGENET1K_V1,
    **kwargs: Any,
) -> FCN:
    """Fully-Convolutional Network model with a ResNet-50 backbone from the `Fully Convolutional
    Networks for Semantic Segmentation <https://arxiv.org/abs/1411.4038>`_ paper.

    .. betastatus:: segmentation module

    Args:
        weights (:class:`~torchvision.models.segmentation.FCN_ResNet50_Weights`, optional): The
            pretrained weights to use. See
            :class:`~torchvision.models.segmentation.FCN_ResNet50_Weights` below for
            more details, and possible values. By default, no pre-trained
            weights are used.
        progress (bool, optional): If True, displays a progress bar of the
            download to stderr. Default is True.
        num_classes (int, optional): number of output classes of the model (including the background).
        aux_loss (bool, optional): If True, it uses an auxiliary loss.
        weights_backbone (:class:`~torchvision.models.ResNet50_Weights`, optional): The pretrained
            weights for the backbone.
        **kwargs: parameters passed to the ``torchvision.models.segmentation.fcn.FCN``
            base class. Please refer to the `source code
            <https://github.com/pytorch/vision/blob/main/torchvision/models/segmentation/fcn.py>`_
            for more details about this class.

    .. autoclass:: torchvision.models.segmentation.FCN_ResNet50_Weights
        :members:
    """

    weights = FCN_ResNet50_Weights.verify(weights)
    weights_backbone = ResNet50_Weights.verify(weights_backbone)

    if weights is not None:
        weights_backbone = None
        num_classes = _ovewrite_value_param("num_classes", num_classes, len(weights.meta["categories"]))
        aux_loss = _ovewrite_value_param("aux_loss", aux_loss, True)
    elif num_classes is None:
        num_classes = 21

    backbone = resnet50(weights=weights_backbone, replace_stride_with_dilation=[False, True, True])
    model = _fcn_resnet(backbone, num_classes, aux_loss)

    if weights is not None:
        model.load_state_dict(weights.get_state_dict(progress=progress, check_hash=True))

    return model


@register_model()
@handle_legacy_interface(
    weights=("pretrained", FCN_ResNet101_Weights.COCO_WITH_VOC_LABELS_V1),
    weights_backbone=("pretrained_backbone", ResNet101_Weights.IMAGENET1K_V1),
)
def fcn_resnet101(
    *,
    weights: Optional[FCN_ResNet101_Weights] = None,
    progress: bool = True,
    num_classes: Optional[int] = None,
    aux_loss: Optional[bool] = None,
    weights_backbone: Optional[ResNet101_Weights] = ResNet101_Weights.IMAGENET1K_V1,
    **kwargs: Any,
) -> FCN:
    """Fully-Convolutional Network model with a ResNet-101 backbone from the `Fully Convolutional
    Networks for Semantic Segmentation <https://arxiv.org/abs/1411.4038>`_ paper.

    .. betastatus:: segmentation module

    Args:
        weights (:class:`~torchvision.models.segmentation.FCN_ResNet101_Weights`, optional): The
            pretrained weights to use. See
            :class:`~torchvision.models.segmentation.FCN_ResNet101_Weights` below for
            more details, and possible values. By default, no pre-trained
            weights are used.
        progress (bool, optional): If True, displays a progress bar of the
            download to stderr. Default is True.
        num_classes (int, optional): number of output classes of the model (including the background).
        aux_loss (bool, optional): If True, it uses an auxiliary loss.
        weights_backbone (:class:`~torchvision.models.ResNet101_Weights`, optional): The pretrained
            weights for the backbone.
        **kwargs: parameters passed to the ``torchvision.models.segmentation.fcn.FCN``
            base class. Please refer to the `source code
            <https://github.com/pytorch/vision/blob/main/torchvision/models/segmentation/fcn.py>`_
            for more details about this class.

    .. autoclass:: torchvision.models.segmentation.FCN_ResNet101_Weights
        :members:
    """

    weights = FCN_ResNet101_Weights.verify(weights)
    weights_backbone = ResNet101_Weights.verify(weights_backbone)

    if weights is not None:
        weights_backbone = None
        num_classes = _ovewrite_value_param("num_classes", num_classes, len(weights.meta["categories"]))
        aux_loss = _ovewrite_value_param("aux_loss", aux_loss, True)
    elif num_classes is None:
        num_classes = 21

    backbone = resnet101(weights=weights_backbone, replace_stride_with_dilation=[False, True, True])
    model = _fcn_resnet(backbone, num_classes, aux_loss)

    if weights is not None:
        model.load_state_dict(weights.get_state_dict(progress=progress, check_hash=True))

    return model



from ._optical_flow import FlyingChairs, FlyingThings3D, HD1K, KittiFlow, Sintel
from ._stereo_matching import (
    CarlaStereo,
    CREStereo,
    ETH3DStereo,
    FallingThingsStereo,
    InStereo2k,
    Kitti2012Stereo,
    Kitti2015Stereo,
    Middlebury2014Stereo,
    SceneFlowStereo,
    SintelStereo,
)
from .caltech import Caltech101, Caltech256
from .celeba import CelebA
from .cifar import CIFAR10, CIFAR100
from .cityscapes import Cityscapes
from .clevr import CLEVRClassification
from .coco import CocoCaptions, CocoDetection
from .country211 import Country211
from .dtd import DTD
from .eurosat import EuroSAT
from .fakedata import FakeData
from .fer2013 import FER2013
from .fgvc_aircraft import FGVCAircraft
from .flickr import Flickr30k, Flickr8k
from .flowers102 import Flowers102
from .folder import DatasetFolder, ImageFolder
from .food101 import Food101
from .gtsrb import GTSRB
from .hmdb51 import HMDB51
from .imagenet import ImageNet
from .imagenette import Imagenette
from .inaturalist import INaturalist
from .kinetics import Kinetics
from .kitti import Kitti
from .lfw import LFWPairs, LFWPeople
from .lsun import LSUN, LSUNClass
from .mnist import EMNIST, FashionMNIST, KMNIST, MNIST, QMNIST
from .moving_mnist import MovingMNIST
from .omniglot import Omniglot
from .oxford_iiit_pet import OxfordIIITPet
from .pcam import PCAM
from .phototour import PhotoTour
from .places365 import Places365
from .rendered_sst2 import RenderedSST2
from .sbd import SBDataset
from .sbu import SBU
from .semeion import SEMEION
from .stanford_cars import StanfordCars
from .stl10 import STL10
from .sun397 import SUN397
from .svhn import SVHN
from .ucf101 import UCF101
from .usps import USPS
from .vision import VisionDataset
from .voc import VOCDetection, VOCSegmentation
from .widerface import WIDERFace

__all__ = (
    "LSUN",
    "LSUNClass",
    "ImageFolder",
    "DatasetFolder",
    "FakeData",
    "CocoCaptions",
    "CocoDetection",
    "CIFAR10",
    "CIFAR100",
    "EMNIST",
    "FashionMNIST",
    "QMNIST",
    "MNIST",
    "KMNIST",
    "StanfordCars",
    "STL10",
    "SUN397",
    "SVHN",
    "PhotoTour",
    "SEMEION",
    "Omniglot",
    "SBU",
    "Flickr8k",
    "Flickr30k",
    "Flowers102",
    "VOCSegmentation",
    "VOCDetection",
    "Cityscapes",
    "ImageNet",
    "Caltech101",
    "Caltech256",
    "CelebA",
    "WIDERFace",
    "SBDataset",
    "VisionDataset",
    "USPS",
    "Kinetics",
    "HMDB51",
    "UCF101",
    "Places365",
    "Kitti",
    "INaturalist",
    "LFWPeople",
    "LFWPairs",
    "KittiFlow",
    "Sintel",
    "FlyingChairs",
    "FlyingThings3D",
    "HD1K",
    "Food101",
    "DTD",
    "FER2013",
    "GTSRB",
    "CLEVRClassification",
    "OxfordIIITPet",
    "PCAM",
    "Country211",
    "FGVCAircraft",
    "EuroSAT",
    "RenderedSST2",
    "Kitti2012Stereo",
    "Kitti2015Stereo",
    "CarlaStereo",
    "Middlebury2014Stereo",
    "CREStereo",
    "FallingThingsStereo",
    "SceneFlowStereo",
    "SintelStereo",
    "InStereo2k",
    "ETH3DStereo",
    "wrap_dataset_for_transforms_v2",
    "Imagenette",
)


# We override current module's attributes to handle the import:
# from torchvision.datasets import wrap_dataset_for_transforms_v2
# without a cyclic error.
# Ref: https://peps.python.org/pep-0562/
def __getattr__(name):
    if name in ("wrap_dataset_for_transforms_v2",):
        from torchvision.tv_tensors._dataset_wrapper import wrap_dataset_for_transforms_v2

        return wrap_dataset_for_transforms_v2

    raise AttributeError(f"module {__name__!r} has no attribute {name!r}")



import warnings
from typing import Any, Dict, Union

import numpy as np
import PIL.Image
import torch
from torchvision.transforms import functional as _F

from torchvision.transforms.v2 import Transform


class ToTensor(Transform):
    """[DEPRECATED] Use ``v2.Compose([v2.ToImage(), v2.ToDtype(torch.float32, scale=True)])`` instead.

    Convert a PIL Image or ndarray to tensor and scale the values accordingly.

    .. warning::
        :class:`v2.ToTensor` is deprecated and will be removed in a future release.
        Please use instead ``v2.Compose([v2.ToImage(), v2.ToDtype(torch.float32, scale=True)])``.
        Output is equivalent up to float precision.

    This transform does not support torchscript.


    Converts a PIL Image or numpy.ndarray (H x W x C) in the range
    [0, 255] to a torch.FloatTensor of shape (C x H x W) in the range [0.0, 1.0]
    if the PIL Image belongs to one of the modes (L, LA, P, I, F, RGB, YCbCr, RGBA, CMYK, 1)
    or if the numpy.ndarray has dtype = np.uint8

    In the other cases, tensors are returned without scaling.

    .. note::
        Because the input image is scaled to [0.0, 1.0], this transformation should not be used when
        transforming target image masks. See the `references`_ for implementing the transforms for image masks.

    .. _references: https://github.com/pytorch/vision/tree/main/references/segmentation
    """

    _transformed_types = (PIL.Image.Image, np.ndarray)

    def __init__(self) -> None:
        warnings.warn(
            "The transform `ToTensor()` is deprecated and will be removed in a future release. "
            "Instead, please use `v2.Compose([v2.ToImage(), v2.ToDtype(torch.float32, scale=True)])`."
            "Output is equivalent up to float precision."
        )
        super().__init__()

    def _transform(self, inpt: Union[PIL.Image.Image, np.ndarray], params: Dict[str, Any]) -> torch.Tensor:
        return _F.to_tensor(inpt)



import math
import os
import random
import re
import sys
from functools import partial

import numpy as np
import pytest
import torch
import torchvision.transforms as transforms
import torchvision.transforms._functional_tensor as F_t
import torchvision.transforms.functional as F
from PIL import Image
from torch._utils_internal import get_file_path_2

try:
    import accimage
except ImportError:
    accimage = None

try:
    from scipy import stats
except ImportError:
    stats = None

from common_utils import assert_equal, cycle_over, float_dtypes, int_dtypes


GRACE_HOPPER = get_file_path_2(
    os.path.dirname(os.path.abspath(__file__)), "assets", "encode_jpeg", "grace_hopper_517x606.jpg"
)


def _get_grayscale_test_image(img, fill=None):
    img = img.convert("L")
    fill = (fill[0],) if isinstance(fill, tuple) else fill
    return img, fill


class TestConvertImageDtype:
    @pytest.mark.parametrize("input_dtype, output_dtype", cycle_over(float_dtypes()))
    def test_float_to_float(self, input_dtype, output_dtype):
        input_image = torch.tensor((0.0, 1.0), dtype=input_dtype)
        transform = transforms.ConvertImageDtype(output_dtype)
        transform_script = torch.jit.script(F.convert_image_dtype)

        output_image = transform(input_image)
        output_image_script = transform_script(input_image, output_dtype)

        torch.testing.assert_close(output_image_script, output_image, rtol=0.0, atol=1e-6)

        actual_min, actual_max = output_image.tolist()
        desired_min, desired_max = 0.0, 1.0

        assert abs(actual_min - desired_min) < 1e-7
        assert abs(actual_max - desired_max) < 1e-7

    @pytest.mark.parametrize("input_dtype", float_dtypes())
    @pytest.mark.parametrize("output_dtype", int_dtypes())
    def test_float_to_int(self, input_dtype, output_dtype):
        input_image = torch.tensor((0.0, 1.0), dtype=input_dtype)
        transform = transforms.ConvertImageDtype(output_dtype)
        transform_script = torch.jit.script(F.convert_image_dtype)

        if (input_dtype == torch.float32 and output_dtype in (torch.int32, torch.int64)) or (
            input_dtype == torch.float64 and output_dtype == torch.int64
        ):
            with pytest.raises(RuntimeError):
                transform(input_image)
        else:
            output_image = transform(input_image)
            output_image_script = transform_script(input_image, output_dtype)

            torch.testing.assert_close(output_image_script, output_image, rtol=0.0, atol=1e-6)

            actual_min, actual_max = output_image.tolist()
            desired_min, desired_max = 0, torch.iinfo(output_dtype).max

            assert actual_min == desired_min
            assert actual_max == desired_max

    @pytest.mark.parametrize("input_dtype", int_dtypes())
    @pytest.mark.parametrize("output_dtype", float_dtypes())
    def test_int_to_float(self, input_dtype, output_dtype):
        input_image = torch.tensor((0, torch.iinfo(input_dtype).max), dtype=input_dtype)
        transform = transforms.ConvertImageDtype(output_dtype)
        transform_script = torch.jit.script(F.convert_image_dtype)

        output_image = transform(input_image)
        output_image_script = transform_script(input_image, output_dtype)

        torch.testing.assert_close(output_image_script, output_image, rtol=0.0, atol=1e-6)

        actual_min, actual_max = output_image.tolist()
        desired_min, desired_max = 0.0, 1.0

        assert abs(actual_min - desired_min) < 1e-7
        assert actual_min >= desired_min
        assert abs(actual_max - desired_max) < 1e-7
        assert actual_max <= desired_max

    @pytest.mark.parametrize("input_dtype, output_dtype", cycle_over(int_dtypes()))
    def test_dtype_int_to_int(self, input_dtype, output_dtype):
        input_max = torch.iinfo(input_dtype).max
        input_image = torch.tensor((0, input_max), dtype=input_dtype)
        output_max = torch.iinfo(output_dtype).max

        transform = transforms.ConvertImageDtype(output_dtype)
        transform_script = torch.jit.script(F.convert_image_dtype)

        output_image = transform(input_image)
        output_image_script = transform_script(input_image, output_dtype)

        torch.testing.assert_close(
            output_image_script,
            output_image,
            rtol=0.0,
            atol=1e-6,
            msg=f"{output_image_script} vs {output_image}",
        )

        actual_min, actual_max = output_image.tolist()
        desired_min, desired_max = 0, output_max

        # see https://github.com/pytorch/vision/pull/2078#issuecomment-641036236 for details
        if input_max >= output_max:
            error_term = 0
        else:
            error_term = 1 - (torch.iinfo(output_dtype).max + 1) // (torch.iinfo(input_dtype).max + 1)

        assert actual_min == desired_min
        assert actual_max == (desired_max + error_term)

    @pytest.mark.parametrize("input_dtype, output_dtype", cycle_over(int_dtypes()))
    def test_int_to_int_consistency(self, input_dtype, output_dtype):
        input_max = torch.iinfo(input_dtype).max
        input_image = torch.tensor((0, input_max), dtype=input_dtype)

        output_max = torch.iinfo(output_dtype).max
        if output_max <= input_max:
            return

        transform = transforms.ConvertImageDtype(output_dtype)
        inverse_transfrom = transforms.ConvertImageDtype(input_dtype)
        output_image = inverse_transfrom(transform(input_image))

        actual_min, actual_max = output_image.tolist()
        desired_min, desired_max = 0, input_max

        assert actual_min == desired_min
        assert actual_max == desired_max


@pytest.mark.skipif(accimage is None, reason="accimage not available")
class TestAccImage:
    def test_accimage_to_tensor(self):
        trans = transforms.PILToTensor()

        expected_output = trans(Image.open(GRACE_HOPPER).convert("RGB"))
        output = trans(accimage.Image(GRACE_HOPPER))

        torch.testing.assert_close(output, expected_output)

    def test_accimage_pil_to_tensor(self):
        trans = transforms.PILToTensor()

        expected_output = trans(Image.open(GRACE_HOPPER).convert("RGB"))
        output = trans(accimage.Image(GRACE_HOPPER))

        assert expected_output.size() == output.size()
        torch.testing.assert_close(output, expected_output)

    def test_accimage_resize(self):
        trans = transforms.Compose(
            [
                transforms.Resize(256, interpolation=Image.LINEAR),
                transforms.PILToTensor(),
                transforms.ConvertImageDtype(dtype=torch.float),
            ]
        )

        # Checking if Compose, Resize and ToTensor can be printed as string
        trans.__repr__()

        expected_output = trans(Image.open(GRACE_HOPPER).convert("RGB"))
        output = trans(accimage.Image(GRACE_HOPPER))

        assert expected_output.size() == output.size()
        assert np.abs((expected_output - output).mean()) < 1e-3
        assert (expected_output - output).var() < 1e-5
        # note the high absolute tolerance
        torch.testing.assert_close(output.numpy(), expected_output.numpy(), rtol=1e-5, atol=5e-2)

    def test_accimage_crop(self):
        trans = transforms.Compose(
            [transforms.CenterCrop(256), transforms.PILToTensor(), transforms.ConvertImageDtype(dtype=torch.float)]
        )

        # Checking if Compose, CenterCrop and ToTensor can be printed as string
        trans.__repr__()

        expected_output = trans(Image.open(GRACE_HOPPER).convert("RGB"))
        output = trans(accimage.Image(GRACE_HOPPER))

        assert expected_output.size() == output.size()
        torch.testing.assert_close(output, expected_output)


class TestToTensor:
    @pytest.mark.parametrize("channels", [1, 3, 4])
    def test_to_tensor(self, channels):
        height, width = 4, 4
        trans = transforms.ToTensor()
        np_rng = np.random.RandomState(0)

        input_data = torch.ByteTensor(channels, height, width).random_(0, 255).float().div_(255)
        img = transforms.ToPILImage()(input_data)
        output = trans(img)
        torch.testing.assert_close(output, input_data)

        ndarray = np_rng.randint(low=0, high=255, size=(height, width, channels)).astype(np.uint8)
        output = trans(ndarray)
        expected_output = ndarray.transpose((2, 0, 1)) / 255.0
        torch.testing.assert_close(output.numpy(), expected_output, check_dtype=False)

        ndarray = np_rng.rand(height, width, channels).astype(np.float32)
        output = trans(ndarray)
        expected_output = ndarray.transpose((2, 0, 1))
        torch.testing.assert_close(output.numpy(), expected_output, check_dtype=False)

        # separate test for mode '1' PIL images
        input_data = torch.ByteTensor(1, height, width).bernoulli_()
        img = transforms.ToPILImage()(input_data.mul(255)).convert("1")
        output = trans(img)
        torch.testing.assert_close(input_data, output, check_dtype=False)

    def test_to_tensor_errors(self):
        height, width = 4, 4
        trans = transforms.ToTensor()
        np_rng = np.random.RandomState(0)

        with pytest.raises(TypeError):
            trans(np_rng.rand(1, height, width).tolist())

        with pytest.raises(ValueError):
            trans(np_rng.rand(height))

        with pytest.raises(ValueError):
            trans(np_rng.rand(1, 1, height, width))

    @pytest.mark.parametrize("dtype", [torch.float16, torch.float, torch.double])
    def test_to_tensor_with_other_default_dtypes(self, dtype):
        np_rng = np.random.RandomState(0)
        current_def_dtype = torch.get_default_dtype()

        t = transforms.ToTensor()
        np_arr = np_rng.randint(0, 255, (32, 32, 3), dtype=np.uint8)
        img = Image.fromarray(np_arr)

        torch.set_default_dtype(dtype)
        res = t(img)
        assert res.dtype == dtype, f"{res.dtype} vs {dtype}"

        torch.set_default_dtype(current_def_dtype)

    @pytest.mark.parametrize("channels", [1, 3, 4])
    def test_pil_to_tensor(self, channels):
        height, width = 4, 4
        trans = transforms.PILToTensor()
        np_rng = np.random.RandomState(0)

        input_data = torch.ByteTensor(channels, height, width).random_(0, 255)
        img = transforms.ToPILImage()(input_data)
        output = trans(img)
        torch.testing.assert_close(input_data, output)

        input_data = np_rng.randint(low=0, high=255, size=(height, width, channels)).astype(np.uint8)
        img = transforms.ToPILImage()(input_data)
        output = trans(img)
        expected_output = input_data.transpose((2, 0, 1))
        torch.testing.assert_close(output.numpy(), expected_output)

        input_data = torch.as_tensor(np_rng.rand(channels, height, width).astype(np.float32))
        img = transforms.ToPILImage()(input_data)  # CHW -> HWC and (* 255).byte()
        output = trans(img)  # HWC -> CHW
        expected_output = (input_data * 255).byte()
        torch.testing.assert_close(output, expected_output)

        # separate test for mode '1' PIL images
        input_data = torch.ByteTensor(1, height, width).bernoulli_()
        img = transforms.ToPILImage()(input_data.mul(255)).convert("1")
        output = trans(img).view(torch.uint8).bool().to(torch.uint8)
        torch.testing.assert_close(input_data, output)

    def test_pil_to_tensor_errors(self):
        height, width = 4, 4
        trans = transforms.PILToTensor()
        np_rng = np.random.RandomState(0)

        with pytest.raises(TypeError):
            trans(np_rng.rand(1, height, width).tolist())

        with pytest.raises(TypeError):
            trans(np_rng.rand(1, height, width))


def test_randomresized_params():
    height = random.randint(24, 32) * 2
    width = random.randint(24, 32) * 2
    img = torch.ones(3, height, width)
    to_pil_image = transforms.ToPILImage()
    img = to_pil_image(img)
    size = 100
    epsilon = 0.05
    min_scale = 0.25
    for _ in range(10):
        scale_min = max(round(random.random(), 2), min_scale)
        scale_range = (scale_min, scale_min + round(random.random(), 2))
        aspect_min = max(round(random.random(), 2), epsilon)
        aspect_ratio_range = (aspect_min, aspect_min + round(random.random(), 2))
        randresizecrop = transforms.RandomResizedCrop(size, scale_range, aspect_ratio_range, antialias=True)
        i, j, h, w = randresizecrop.get_params(img, scale_range, aspect_ratio_range)
        aspect_ratio_obtained = w / h
        assert (
            min(aspect_ratio_range) - epsilon <= aspect_ratio_obtained
            and aspect_ratio_obtained <= max(aspect_ratio_range) + epsilon
        ) or aspect_ratio_obtained == 1.0
        assert isinstance(i, int)
        assert isinstance(j, int)
        assert isinstance(h, int)
        assert isinstance(w, int)


@pytest.mark.parametrize(
    "height, width",
    [
        # height, width
        # square image
        (28, 28),
        (27, 27),
        # rectangular image: h < w
        (28, 34),
        (29, 35),
        # rectangular image: h > w
        (34, 28),
        (35, 29),
    ],
)
@pytest.mark.parametrize(
    "osize",
    [
        # single integer
        22,
        27,
        28,
        36,
        # single integer in tuple/list
        [
            22,
        ],
        (27,),
    ],
)
@pytest.mark.parametrize("max_size", (None, 37, 1000))
def test_resize(height, width, osize, max_size):
    img = Image.new("RGB", size=(width, height), color=127)

    t = transforms.Resize(osize, max_size=max_size, antialias=True)
    result = t(img)

    msg = f"{height}, {width} - {osize} - {max_size}"
    osize = osize[0] if isinstance(osize, (list, tuple)) else osize
    # If size is an int, smaller edge of the image will be matched to this number.
    # i.e, if height > width, then image will be rescaled to (size * height / width, size).
    if height < width:
        exp_w, exp_h = (int(osize * width / height), osize)  # (w, h)
        if max_size is not None and max_size < exp_w:
            exp_w, exp_h = max_size, int(max_size * exp_h / exp_w)
        assert result.size == (exp_w, exp_h), msg
    elif width < height:
        exp_w, exp_h = (osize, int(osize * height / width))  # (w, h)
        if max_size is not None and max_size < exp_h:
            exp_w, exp_h = int(max_size * exp_w / exp_h), max_size
        assert result.size == (exp_w, exp_h), msg
    else:
        exp_w, exp_h = (osize, osize)  # (w, h)
        if max_size is not None and max_size < osize:
            exp_w, exp_h = max_size, max_size
        assert result.size == (exp_w, exp_h), msg


@pytest.mark.parametrize(
    "height, width",
    [
        # height, width
        # square image
        (28, 28),
        (27, 27),
        # rectangular image: h < w
        (28, 34),
        (29, 35),
        # rectangular image: h > w
        (34, 28),
        (35, 29),
    ],
)
@pytest.mark.parametrize(
    "osize",
    [
        # two integers sequence output
        [22, 22],
        [22, 28],
        [22, 36],
        [27, 22],
        [36, 22],
        [28, 28],
        [28, 37],
        [37, 27],
        [37, 37],
    ],
)
def test_resize_sequence_output(height, width, osize):
    img = Image.new("RGB", size=(width, height), color=127)
    oheight, owidth = osize

    t = transforms.Resize(osize, antialias=True)
    result = t(img)

    assert (owidth, oheight) == result.size


def test_resize_antialias_error():
    osize = [37, 37]
    img = Image.new("RGB", size=(35, 29), color=127)

    with pytest.warns(UserWarning, match=r"Anti-alias option is always applied for PIL Image input"):
        t = transforms.Resize(osize, antialias=False)
        t(img)


@pytest.mark.parametrize("height, width", ((32, 64), (64, 32)))
def test_resize_size_equals_small_edge_size(height, width):
    # Non-regression test for https://github.com/pytorch/vision/issues/5405
    # max_size used to be ignored if size == small_edge_size
    max_size = 40
    img = Image.new("RGB", size=(width, height), color=127)

    small_edge = min(height, width)
    t = transforms.Resize(small_edge, max_size=max_size, antialias=True)
    result = t(img)
    assert max(result.size) == max_size


def test_resize_equal_input_output_sizes():
    # Regression test for https://github.com/pytorch/vision/issues/7518
    height, width = 28, 27
    img = Image.new("RGB", size=(width, height))

    t = transforms.Resize((height, width), antialias=True)
    result = t(img)
    assert result is img


class TestPad:
    @pytest.mark.parametrize("fill", [85, 85.0])
    def test_pad(self, fill):
        height = random.randint(10, 32) * 2
        width = random.randint(10, 32) * 2
        img = torch.ones(3, height, width, dtype=torch.uint8)
        padding = random.randint(1, 20)
        result = transforms.Compose(
            [
                transforms.ToPILImage(),
                transforms.Pad(padding, fill=fill),
                transforms.PILToTensor(),
            ]
        )(img)
        assert result.size(1) == height + 2 * padding
        assert result.size(2) == width + 2 * padding
        # check that all elements in the padded region correspond
        # to the pad value
        h_padded = result[:, :padding, :]
        w_padded = result[:, :, :padding]
        torch.testing.assert_close(h_padded, torch.full_like(h_padded, fill_value=fill), rtol=0.0, atol=0.0)
        torch.testing.assert_close(w_padded, torch.full_like(w_padded, fill_value=fill), rtol=0.0, atol=0.0)
        pytest.raises(ValueError, transforms.Pad(padding, fill=(1, 2)), transforms.ToPILImage()(img))

    def test_pad_with_tuple_of_pad_values(self):
        height = random.randint(10, 32) * 2
        width = random.randint(10, 32) * 2
        img = transforms.ToPILImage()(torch.ones(3, height, width))

        padding = tuple(random.randint(1, 20) for _ in range(2))
        output = transforms.Pad(padding)(img)
        assert output.size == (width + padding[0] * 2, height + padding[1] * 2)

        padding = [random.randint(1, 20) for _ in range(4)]
        output = transforms.Pad(padding)(img)
        assert output.size[0] == width + padding[0] + padding[2]
        assert output.size[1] == height + padding[1] + padding[3]

        # Checking if Padding can be printed as string
        transforms.Pad(padding).__repr__()

    def test_pad_with_non_constant_padding_modes(self):
        """Unit tests for edge, reflect, symmetric padding"""
        img = torch.zeros(3, 27, 27).byte()
        img[:, :, 0] = 1  # Constant value added to leftmost edge
        img = transforms.ToPILImage()(img)
        img = F.pad(img, 1, (200, 200, 200))

        # pad 3 to all sidess
        edge_padded_img = F.pad(img, 3, padding_mode="edge")
        # First 6 elements of leftmost edge in the middle of the image, values are in order:
        # edge_pad, edge_pad, edge_pad, constant_pad, constant value added to leftmost edge, 0
        edge_middle_slice = np.asarray(edge_padded_img).transpose(2, 0, 1)[0][17][:6]
        assert_equal(edge_middle_slice, np.asarray([200, 200, 200, 200, 1, 0], dtype=np.uint8))
        assert transforms.PILToTensor()(edge_padded_img).size() == (3, 35, 35)

        # Pad 3 to left/right, 2 to top/bottom
        reflect_padded_img = F.pad(img, (3, 2), padding_mode="reflect")
        # First 6 elements of leftmost edge in the middle of the image, values are in order:
        # reflect_pad, reflect_pad, reflect_pad, constant_pad, constant value added to leftmost edge, 0
        reflect_middle_slice = np.asarray(reflect_padded_img).transpose(2, 0, 1)[0][17][:6]
        assert_equal(reflect_middle_slice, np.asarray([0, 0, 1, 200, 1, 0], dtype=np.uint8))
        assert transforms.PILToTensor()(reflect_padded_img).size() == (3, 33, 35)

        # Pad 3 to left, 2 to top, 2 to right, 1 to bottom
        symmetric_padded_img = F.pad(img, (3, 2, 2, 1), padding_mode="symmetric")
        # First 6 elements of leftmost edge in the middle of the image, values are in order:
        # sym_pad, sym_pad, sym_pad, constant_pad, constant value added to leftmost edge, 0
        symmetric_middle_slice = np.asarray(symmetric_padded_img).transpose(2, 0, 1)[0][17][:6]
        assert_equal(symmetric_middle_slice, np.asarray([0, 1, 200, 200, 1, 0], dtype=np.uint8))
        assert transforms.PILToTensor()(symmetric_padded_img).size() == (3, 32, 34)

        # Check negative padding explicitly for symmetric case, since it is not
        # implemented for tensor case to compare to
        # Crop 1 to left, pad 2 to top, pad 3 to right, crop 3 to bottom
        symmetric_padded_img_neg = F.pad(img, (-1, 2, 3, -3), padding_mode="symmetric")
        symmetric_neg_middle_left = np.asarray(symmetric_padded_img_neg).transpose(2, 0, 1)[0][17][:3]
        symmetric_neg_middle_right = np.asarray(symmetric_padded_img_neg).transpose(2, 0, 1)[0][17][-4:]
        assert_equal(symmetric_neg_middle_left, np.asarray([1, 0, 0], dtype=np.uint8))
        assert_equal(symmetric_neg_middle_right, np.asarray([200, 200, 0, 0], dtype=np.uint8))
        assert transforms.PILToTensor()(symmetric_padded_img_neg).size() == (3, 28, 31)

    def test_pad_raises_with_invalid_pad_sequence_len(self):
        with pytest.raises(ValueError):
            transforms.Pad(())

        with pytest.raises(ValueError):
            transforms.Pad((1, 2, 3))

        with pytest.raises(ValueError):
            transforms.Pad((1, 2, 3, 4, 5))

    def test_pad_with_mode_F_images(self):
        pad = 2
        transform = transforms.Pad(pad)

        img = Image.new("F", (10, 10))
        padded_img = transform(img)
        assert_equal(padded_img.size, [edge_size + 2 * pad for edge_size in img.size])


@pytest.mark.parametrize(
    "fn, trans, kwargs",
    [
        (F.invert, transforms.RandomInvert, {}),
        (F.posterize, transforms.RandomPosterize, {"bits": 4}),
        (F.solarize, transforms.RandomSolarize, {"threshold": 192}),
        (F.adjust_sharpness, transforms.RandomAdjustSharpness, {"sharpness_factor": 2.0}),
        (F.autocontrast, transforms.RandomAutocontrast, {}),
        (F.equalize, transforms.RandomEqualize, {}),
        (F.vflip, transforms.RandomVerticalFlip, {}),
        (F.hflip, transforms.RandomHorizontalFlip, {}),
        (partial(F.to_grayscale, num_output_channels=3), transforms.RandomGrayscale, {}),
    ],
)
@pytest.mark.parametrize("seed", range(10))
@pytest.mark.parametrize("p", (0, 1))
def test_randomness(fn, trans, kwargs, seed, p):
    torch.manual_seed(seed)
    img = transforms.ToPILImage()(torch.rand(3, 16, 18))

    expected_transformed_img = fn(img, **kwargs)
    randomly_transformed_img = trans(p=p, **kwargs)(img)

    if p == 0:
        assert randomly_transformed_img == img
    elif p == 1:
        assert randomly_transformed_img == expected_transformed_img

    trans(**kwargs).__repr__()


def test_autocontrast_equal_minmax():
    img_tensor = torch.tensor([[[10]], [[128]], [[245]]], dtype=torch.uint8).expand(3, 32, 32)
    img_pil = F.to_pil_image(img_tensor)

    img_tensor = F.autocontrast(img_tensor)
    img_pil = F.autocontrast(img_pil)
    torch.testing.assert_close(img_tensor, F.pil_to_tensor(img_pil))


class TestToPil:
    def _get_1_channel_tensor_various_types():
        img_data_float = torch.Tensor(1, 4, 4).uniform_()
        expected_output = img_data_float.mul(255).int().float().div(255).numpy()
        yield img_data_float, expected_output, "L"

        img_data_byte = torch.ByteTensor(1, 4, 4).random_(0, 255)
        expected_output = img_data_byte.float().div(255.0).numpy()
        yield img_data_byte, expected_output, "L"

        img_data_short = torch.ShortTensor(1, 4, 4).random_()
        expected_output = img_data_short.numpy()
        yield img_data_short, expected_output, "I;16" if sys.byteorder == "little" else "I;16B"

        img_data_int = torch.IntTensor(1, 4, 4).random_()
        expected_output = img_data_int.numpy()
        yield img_data_int, expected_output, "I"

    def _get_2d_tensor_various_types():
        img_data_float = torch.Tensor(4, 4).uniform_()
        expected_output = img_data_float.mul(255).int().float().div(255).numpy()
        yield img_data_float, expected_output, "L"

        img_data_byte = torch.ByteTensor(4, 4).random_(0, 255)
        expected_output = img_data_byte.float().div(255.0).numpy()
        yield img_data_byte, expected_output, "L"

        img_data_short = torch.ShortTensor(4, 4).random_()
        expected_output = img_data_short.numpy()
        yield img_data_short, expected_output, "I;16" if sys.byteorder == "little" else "I;16B"

        img_data_int = torch.IntTensor(4, 4).random_()
        expected_output = img_data_int.numpy()
        yield img_data_int, expected_output, "I"

    @pytest.mark.parametrize("with_mode", [False, True])
    @pytest.mark.parametrize("img_data, expected_output, expected_mode", _get_1_channel_tensor_various_types())
    def test_1_channel_tensor_to_pil_image(self, with_mode, img_data, expected_output, expected_mode):
        transform = transforms.ToPILImage(mode=expected_mode) if with_mode else transforms.ToPILImage()
        to_tensor = transforms.ToTensor()

        img = transform(img_data)
        assert img.mode == expected_mode
        torch.testing.assert_close(expected_output, to_tensor(img).numpy())

    def test_1_channel_float_tensor_to_pil_image(self):
        img_data = torch.Tensor(1, 4, 4).uniform_()
        # 'F' mode for torch.FloatTensor
        img_F_mode = transforms.ToPILImage(mode="F")(img_data)
        assert img_F_mode.mode == "F"
        torch.testing.assert_close(
            np.array(Image.fromarray(img_data.squeeze(0).numpy(), mode="F")), np.array(img_F_mode)
        )

    @pytest.mark.parametrize("with_mode", [False, True])
    @pytest.mark.parametrize(
        "img_data, expected_mode",
        [
            (torch.Tensor(4, 4, 1).uniform_().numpy(), "L"),
            (torch.ByteTensor(4, 4, 1).random_(0, 255).numpy(), "L"),
            (torch.ShortTensor(4, 4, 1).random_().numpy(), "I;16" if sys.byteorder == "little" else "I;16B"),
            (torch.IntTensor(4, 4, 1).random_().numpy(), "I"),
        ],
    )
    def test_1_channel_ndarray_to_pil_image(self, with_mode, img_data, expected_mode):
        transform = transforms.ToPILImage(mode=expected_mode) if with_mode else transforms.ToPILImage()
        img = transform(img_data)
        assert img.mode == expected_mode
        if np.issubdtype(img_data.dtype, np.floating):
            img_data = (img_data * 255).astype(np.uint8)
        # note: we explicitly convert img's dtype because pytorch doesn't support uint16
        # and otherwise assert_close wouldn't be able to construct a tensor from the uint16 array
        torch.testing.assert_close(img_data[:, :, 0], np.asarray(img).astype(img_data.dtype))

    @pytest.mark.parametrize("expected_mode", [None, "LA"])
    def test_2_channel_ndarray_to_pil_image(self, expected_mode):
        img_data = torch.ByteTensor(4, 4, 2).random_(0, 255).numpy()

        if expected_mode is None:
            img = transforms.ToPILImage()(img_data)
            assert img.mode == "LA"  # default should assume LA
        else:
            img = transforms.ToPILImage(mode=expected_mode)(img_data)
            assert img.mode == expected_mode
        split = img.split()
        for i in range(2):
            torch.testing.assert_close(img_data[:, :, i], np.asarray(split[i]))

    def test_2_channel_ndarray_to_pil_image_error(self):
        img_data = torch.ByteTensor(4, 4, 2).random_(0, 255).numpy()
        transforms.ToPILImage().__repr__()

        # should raise if we try a mode for 4 or 1 or 3 channel images
        with pytest.raises(ValueError, match=r"Only modes \['LA'\] are supported for 2D inputs"):
            transforms.ToPILImage(mode="RGBA")(img_data)
        with pytest.raises(ValueError, match=r"Only modes \['LA'\] are supported for 2D inputs"):
            transforms.ToPILImage(mode="P")(img_data)
        with pytest.raises(ValueError, match=r"Only modes \['LA'\] are supported for 2D inputs"):
            transforms.ToPILImage(mode="RGB")(img_data)

    @pytest.mark.parametrize("expected_mode", [None, "LA"])
    def test_2_channel_tensor_to_pil_image(self, expected_mode):
        img_data = torch.Tensor(2, 4, 4).uniform_()
        expected_output = img_data.mul(255).int().float().div(255)
        if expected_mode is None:
            img = transforms.ToPILImage()(img_data)
            assert img.mode == "LA"  # default should assume LA
        else:
            img = transforms.ToPILImage(mode=expected_mode)(img_data)
            assert img.mode == expected_mode

        split = img.split()
        for i in range(2):
            torch.testing.assert_close(expected_output[i].numpy(), F.to_tensor(split[i]).squeeze(0).numpy())

    def test_2_channel_tensor_to_pil_image_error(self):
        img_data = torch.Tensor(2, 4, 4).uniform_()

        # should raise if we try a mode for 4 or 1 or 3 channel images
        with pytest.raises(ValueError, match=r"Only modes \['LA'\] are supported for 2D inputs"):
            transforms.ToPILImage(mode="RGBA")(img_data)
        with pytest.raises(ValueError, match=r"Only modes \['LA'\] are supported for 2D inputs"):
            transforms.ToPILImage(mode="P")(img_data)
        with pytest.raises(ValueError, match=r"Only modes \['LA'\] are supported for 2D inputs"):
            transforms.ToPILImage(mode="RGB")(img_data)

    @pytest.mark.parametrize("with_mode", [False, True])
    @pytest.mark.parametrize("img_data, expected_output, expected_mode", _get_2d_tensor_various_types())
    def test_2d_tensor_to_pil_image(self, with_mode, img_data, expected_output, expected_mode):
        transform = transforms.ToPILImage(mode=expected_mode) if with_mode else transforms.ToPILImage()
        to_tensor = transforms.ToTensor()

        img = transform(img_data)
        assert img.mode == expected_mode
        torch.testing.assert_close(expected_output, to_tensor(img).numpy()[0])

    @pytest.mark.parametrize("with_mode", [False, True])
    @pytest.mark.parametrize(
        "img_data, expected_mode",
        [
            (torch.Tensor(4, 4).uniform_().numpy(), "L"),
            (torch.ByteTensor(4, 4).random_(0, 255).numpy(), "L"),
            (torch.ShortTensor(4, 4).random_().numpy(), "I;16" if sys.byteorder == "little" else "I;16B"),
            (torch.IntTensor(4, 4).random_().numpy(), "I"),
        ],
    )
    def test_2d_ndarray_to_pil_image(self, with_mode, img_data, expected_mode):
        transform = transforms.ToPILImage(mode=expected_mode) if with_mode else transforms.ToPILImage()
        img = transform(img_data)
        assert img.mode == expected_mode
        if np.issubdtype(img_data.dtype, np.floating):
            img_data = (img_data * 255).astype(np.uint8)
        np.testing.assert_allclose(img_data, img)

    @pytest.mark.parametrize("expected_mode", [None, "RGB", "HSV", "YCbCr"])
    def test_3_channel_tensor_to_pil_image(self, expected_mode):
        img_data = torch.Tensor(3, 4, 4).uniform_()
        expected_output = img_data.mul(255).int().float().div(255)

        if expected_mode is None:
            img = transforms.ToPILImage()(img_data)
            assert img.mode == "RGB"  # default should assume RGB
        else:
            img = transforms.ToPILImage(mode=expected_mode)(img_data)
            assert img.mode == expected_mode
        split = img.split()
        for i in range(3):
            torch.testing.assert_close(expected_output[i].numpy(), F.to_tensor(split[i]).squeeze(0).numpy())

    def test_3_channel_tensor_to_pil_image_error(self):
        img_data = torch.Tensor(3, 4, 4).uniform_()
        error_message_3d = r"Only modes \['RGB', 'YCbCr', 'HSV'\] are supported for 3D inputs"
        # should raise if we try a mode for 4 or 1 or 2 channel images
        with pytest.raises(ValueError, match=error_message_3d):
            transforms.ToPILImage(mode="RGBA")(img_data)
        with pytest.raises(ValueError, match=error_message_3d):
            transforms.ToPILImage(mode="P")(img_data)
        with pytest.raises(ValueError, match=error_message_3d):
            transforms.ToPILImage(mode="LA")(img_data)

        with pytest.raises(ValueError, match=r"pic should be 2/3 dimensional. Got \d+ dimensions."):
            transforms.ToPILImage()(torch.Tensor(1, 3, 4, 4).uniform_())

    @pytest.mark.parametrize("expected_mode", [None, "RGB", "HSV", "YCbCr"])
    def test_3_channel_ndarray_to_pil_image(self, expected_mode):
        img_data = torch.ByteTensor(4, 4, 3).random_(0, 255).numpy()

        if expected_mode is None:
            img = transforms.ToPILImage()(img_data)
            assert img.mode == "RGB"  # default should assume RGB
        else:
            img = transforms.ToPILImage(mode=expected_mode)(img_data)
            assert img.mode == expected_mode
        split = img.split()
        for i in range(3):
            torch.testing.assert_close(img_data[:, :, i], np.asarray(split[i]))

    def test_3_channel_ndarray_to_pil_image_error(self):
        img_data = torch.ByteTensor(4, 4, 3).random_(0, 255).numpy()

        # Checking if ToPILImage can be printed as string
        transforms.ToPILImage().__repr__()

        error_message_3d = r"Only modes \['RGB', 'YCbCr', 'HSV'\] are supported for 3D inputs"
        # should raise if we try a mode for 4 or 1 or 2 channel images
        with pytest.raises(ValueError, match=error_message_3d):
            transforms.ToPILImage(mode="RGBA")(img_data)
        with pytest.raises(ValueError, match=error_message_3d):
            transforms.ToPILImage(mode="P")(img_data)
        with pytest.raises(ValueError, match=error_message_3d):
            transforms.ToPILImage(mode="LA")(img_data)

    @pytest.mark.parametrize("expected_mode", [None, "RGBA", "CMYK", "RGBX"])
    def test_4_channel_tensor_to_pil_image(self, expected_mode):
        img_data = torch.Tensor(4, 4, 4).uniform_()
        expected_output = img_data.mul(255).int().float().div(255)

        if expected_mode is None:
            img = transforms.ToPILImage()(img_data)
            assert img.mode == "RGBA"  # default should assume RGBA
        else:
            img = transforms.ToPILImage(mode=expected_mode)(img_data)
            assert img.mode == expected_mode

        split = img.split()
        for i in range(4):
            torch.testing.assert_close(expected_output[i].numpy(), F.to_tensor(split[i]).squeeze(0).numpy())

    def test_4_channel_tensor_to_pil_image_error(self):
        img_data = torch.Tensor(4, 4, 4).uniform_()

        error_message_4d = r"Only modes \['RGBA', 'CMYK', 'RGBX'\] are supported for 4D inputs"
        # should raise if we try a mode for 3 or 1 or 2 channel images
        with pytest.raises(ValueError, match=error_message_4d):
            transforms.ToPILImage(mode="RGB")(img_data)
        with pytest.raises(ValueError, match=error_message_4d):
            transforms.ToPILImage(mode="P")(img_data)
        with pytest.raises(ValueError, match=error_message_4d):
            transforms.ToPILImage(mode="LA")(img_data)

    @pytest.mark.parametrize("expected_mode", [None, "RGBA", "CMYK", "RGBX"])
    def test_4_channel_ndarray_to_pil_image(self, expected_mode):
        img_data = torch.ByteTensor(4, 4, 4).random_(0, 255).numpy()

        if expected_mode is None:
            img = transforms.ToPILImage()(img_data)
            assert img.mode == "RGBA"  # default should assume RGBA
        else:
            img = transforms.ToPILImage(mode=expected_mode)(img_data)
            assert img.mode == expected_mode
        split = img.split()
        for i in range(4):
            torch.testing.assert_close(img_data[:, :, i], np.asarray(split[i]))

    def test_4_channel_ndarray_to_pil_image_error(self):
        img_data = torch.ByteTensor(4, 4, 4).random_(0, 255).numpy()

        error_message_4d = r"Only modes \['RGBA', 'CMYK', 'RGBX'\] are supported for 4D inputs"
        # should raise if we try a mode for 3 or 1 or 2 channel images
        with pytest.raises(ValueError, match=error_message_4d):
            transforms.ToPILImage(mode="RGB")(img_data)
        with pytest.raises(ValueError, match=error_message_4d):
            transforms.ToPILImage(mode="P")(img_data)
        with pytest.raises(ValueError, match=error_message_4d):
            transforms.ToPILImage(mode="LA")(img_data)

    def test_ndarray_bad_types_to_pil_image(self):
        trans = transforms.ToPILImage()
        reg_msg = r"Input type \w+ is not supported"
        with pytest.raises(TypeError, match=reg_msg):
            trans(np.ones([4, 4, 1], np.int64))
        with pytest.raises(TypeError, match=reg_msg):
            trans(np.ones([4, 4, 1], np.uint16))
        with pytest.raises(TypeError, match=reg_msg):
            trans(np.ones([4, 4, 1], np.uint32))

        with pytest.raises(ValueError, match=r"pic should be 2/3 dimensional. Got \d+ dimensions."):
            transforms.ToPILImage()(np.ones([1, 4, 4, 3]))
        with pytest.raises(ValueError, match=r"pic should not have > 4 channels. Got \d+ channels."):
            transforms.ToPILImage()(np.ones([4, 4, 6]))

    def test_tensor_bad_types_to_pil_image(self):
        with pytest.raises(ValueError, match=r"pic should be 2/3 dimensional. Got \d+ dimensions."):
            transforms.ToPILImage()(torch.ones(1, 3, 4, 4))
        with pytest.raises(ValueError, match=r"pic should not have > 4 channels. Got \d+ channels."):
            transforms.ToPILImage()(torch.ones(6, 4, 4))


def test_adjust_brightness():
    x_shape = [2, 2, 3]
    x_data = [0, 5, 13, 54, 135, 226, 37, 8, 234, 90, 255, 1]
    x_np = np.array(x_data, dtype=np.uint8).reshape(x_shape)
    x_pil = Image.fromarray(x_np, mode="RGB")

    # test 0
    y_pil = F.adjust_brightness(x_pil, 1)
    y_np = np.array(y_pil)
    torch.testing.assert_close(y_np, x_np)

    # test 1
    y_pil = F.adjust_brightness(x_pil, 0.5)
    y_np = np.array(y_pil)
    y_ans = [0, 2, 6, 27, 67, 113, 18, 4, 117, 45, 127, 0]
    y_ans = np.array(y_ans, dtype=np.uint8).reshape(x_shape)
    torch.testing.assert_close(y_np, y_ans)

    # test 2
    y_pil = F.adjust_brightness(x_pil, 2)
    y_np = np.array(y_pil)
    y_ans = [0, 10, 26, 108, 255, 255, 74, 16, 255, 180, 255, 2]
    y_ans = np.array(y_ans, dtype=np.uint8).reshape(x_shape)
    torch.testing.assert_close(y_np, y_ans)


def test_adjust_contrast():
    x_shape = [2, 2, 3]
    x_data = [0, 5, 13, 54, 135, 226, 37, 8, 234, 90, 255, 1]
    x_np = np.array(x_data, dtype=np.uint8).reshape(x_shape)
    x_pil = Image.fromarray(x_np, mode="RGB")

    # test 0
    y_pil = F.adjust_contrast(x_pil, 1)
    y_np = np.array(y_pil)
    torch.testing.assert_close(y_np, x_np)

    # test 1
    y_pil = F.adjust_contrast(x_pil, 0.5)
    y_np = np.array(y_pil)
    y_ans = [43, 45, 49, 70, 110, 156, 61, 47, 160, 88, 170, 43]
    y_ans = np.array(y_ans, dtype=np.uint8).reshape(x_shape)
    torch.testing.assert_close(y_np, y_ans)

    # test 2
    y_pil = F.adjust_contrast(x_pil, 2)
    y_np = np.array(y_pil)
    y_ans = [0, 0, 0, 22, 184, 255, 0, 0, 255, 94, 255, 0]
    y_ans = np.array(y_ans, dtype=np.uint8).reshape(x_shape)
    torch.testing.assert_close(y_np, y_ans)


def test_adjust_hue():
    x_shape = [2, 2, 3]
    x_data = [0, 5, 13, 54, 135, 226, 37, 8, 234, 90, 255, 1]
    x_np = np.array(x_data, dtype=np.uint8).reshape(x_shape)
    x_pil = Image.fromarray(x_np, mode="RGB")

    with pytest.raises(ValueError):
        F.adjust_hue(x_pil, -0.7)
        F.adjust_hue(x_pil, 1)

    # test 0: almost same as x_data but not exact.
    # probably because hsv <-> rgb floating point ops
    y_pil = F.adjust_hue(x_pil, 0)
    y_np = np.array(y_pil)
    y_ans = [0, 5, 13, 54, 139, 226, 35, 8, 234, 91, 255, 1]
    y_ans = np.array(y_ans, dtype=np.uint8).reshape(x_shape)
    torch.testing.assert_close(y_np, y_ans)

    # test 1
    y_pil = F.adjust_hue(x_pil, 0.25)
    y_np = np.array(y_pil)
    y_ans = [13, 0, 12, 224, 54, 226, 234, 8, 99, 1, 222, 255]
    y_ans = np.array(y_ans, dtype=np.uint8).reshape(x_shape)
    torch.testing.assert_close(y_np, y_ans)

    # test 2
    y_pil = F.adjust_hue(x_pil, -0.25)
    y_np = np.array(y_pil)
    y_ans = [0, 13, 2, 54, 226, 58, 8, 234, 152, 255, 43, 1]
    y_ans = np.array(y_ans, dtype=np.uint8).reshape(x_shape)
    torch.testing.assert_close(y_np, y_ans)


def test_adjust_sharpness():
    x_shape = [4, 4, 3]
    x_data = [
        75,
        121,
        114,
        105,
        97,
        107,
        105,
        32,
        66,
        111,
        117,
        114,
        99,
        104,
        97,
        0,
        0,
        65,
        108,
        101,
        120,
        97,
        110,
        100,
        101,
        114,
        32,
        86,
        114,
        121,
        110,
        105,
        111,
        116,
        105,
        115,
        0,
        0,
        73,
        32,
        108,
        111,
        118,
        101,
        32,
        121,
        111,
        117,
    ]
    x_np = np.array(x_data, dtype=np.uint8).reshape(x_shape)
    x_pil = Image.fromarray(x_np, mode="RGB")

    # test 0
    y_pil = F.adjust_sharpness(x_pil, 1)
    y_np = np.array(y_pil)
    torch.testing.assert_close(y_np, x_np)

    # test 1
    y_pil = F.adjust_sharpness(x_pil, 0.5)
    y_np = np.array(y_pil)
    y_ans = [
        75,
        121,
        114,
        105,
        97,
        107,
        105,
        32,
        66,
        111,
        117,
        114,
        99,
        104,
        97,
        30,
        30,
        74,
        103,
        96,
        114,
        97,
        110,
        100,
        101,
        114,
        32,
        81,
        103,
        108,
        102,
        101,
        107,
        116,
        105,
        115,
        0,
        0,
        73,
        32,
        108,
        111,
        118,
        101,
        32,
        121,
        111,
        117,
    ]
    y_ans = np.array(y_ans, dtype=np.uint8).reshape(x_shape)
    torch.testing.assert_close(y_np, y_ans)

    # test 2
    y_pil = F.adjust_sharpness(x_pil, 2)
    y_np = np.array(y_pil)
    y_ans = [
        75,
        121,
        114,
        105,
        97,
        107,
        105,
        32,
        66,
        111,
        117,
        114,
        99,
        104,
        97,
        0,
        0,
        46,
        118,
        111,
        132,
        97,
        110,
        100,
        101,
        114,
        32,
        95,
        135,
        146,
        126,
        112,
        119,
        116,
        105,
        115,
        0,
        0,
        73,
        32,
        108,
        111,
        118,
        101,
        32,
        121,
        111,
        117,
    ]
    y_ans = np.array(y_ans, dtype=np.uint8).reshape(x_shape)
    torch.testing.assert_close(y_np, y_ans)

    # test 3
    x_shape = [2, 2, 3]
    x_data = [0, 5, 13, 54, 135, 226, 37, 8, 234, 90, 255, 1]
    x_np = np.array(x_data, dtype=np.uint8).reshape(x_shape)
    x_pil = Image.fromarray(x_np, mode="RGB")
    x_th = torch.tensor(x_np.transpose(2, 0, 1))
    y_pil = F.adjust_sharpness(x_pil, 2)
    y_np = np.array(y_pil).transpose(2, 0, 1)
    y_th = F.adjust_sharpness(x_th, 2)
    torch.testing.assert_close(y_np, y_th.numpy())


def test_adjust_gamma():
    x_shape = [2, 2, 3]
    x_data = [0, 5, 13, 54, 135, 226, 37, 8, 234, 90, 255, 1]
    x_np = np.array(x_data, dtype=np.uint8).reshape(x_shape)
    x_pil = Image.fromarray(x_np, mode="RGB")

    # test 0
    y_pil = F.adjust_gamma(x_pil, 1)
    y_np = np.array(y_pil)
    torch.testing.assert_close(y_np, x_np)

    # test 1
    y_pil = F.adjust_gamma(x_pil, 0.5)
    y_np = np.array(y_pil)
    y_ans = [0, 35, 57, 117, 186, 241, 97, 45, 245, 152, 255, 16]
    y_ans = np.array(y_ans, dtype=np.uint8).reshape(x_shape)
    torch.testing.assert_close(y_np, y_ans)

    # test 2
    y_pil = F.adjust_gamma(x_pil, 2)
    y_np = np.array(y_pil)
    y_ans = [0, 0, 0, 11, 71, 201, 5, 0, 215, 31, 255, 0]
    y_ans = np.array(y_ans, dtype=np.uint8).reshape(x_shape)
    torch.testing.assert_close(y_np, y_ans)


def test_adjusts_L_mode():
    x_shape = [2, 2, 3]
    x_data = [0, 5, 13, 54, 135, 226, 37, 8, 234, 90, 255, 1]
    x_np = np.array(x_data, dtype=np.uint8).reshape(x_shape)
    x_rgb = Image.fromarray(x_np, mode="RGB")

    x_l = x_rgb.convert("L")
    assert F.adjust_brightness(x_l, 2).mode == "L"
    assert F.adjust_saturation(x_l, 2).mode == "L"
    assert F.adjust_contrast(x_l, 2).mode == "L"
    assert F.adjust_hue(x_l, 0.4).mode == "L"
    assert F.adjust_sharpness(x_l, 2).mode == "L"
    assert F.adjust_gamma(x_l, 0.5).mode == "L"


def test_rotate():
    x = np.zeros((100, 100, 3), dtype=np.uint8)
    x[40, 40] = [255, 255, 255]

    with pytest.raises(TypeError, match=r"img should be PIL Image"):
        F.rotate(x, 10)

    img = F.to_pil_image(x)

    result = F.rotate(img, 45)
    assert result.size == (100, 100)
    r, c, ch = np.where(result)
    assert all(x in r for x in [49, 50])
    assert all(x in c for x in [36])
    assert all(x in ch for x in [0, 1, 2])

    result = F.rotate(img, 45, expand=True)
    assert result.size == (142, 142)
    r, c, ch = np.where(result)
    assert all(x in r for x in [70, 71])
    assert all(x in c for x in [57])
    assert all(x in ch for x in [0, 1, 2])

    result = F.rotate(img, 45, center=(40, 40))
    assert result.size == (100, 100)
    r, c, ch = np.where(result)
    assert all(x in r for x in [40])
    assert all(x in c for x in [40])
    assert all(x in ch for x in [0, 1, 2])

    result_a = F.rotate(img, 90)
    result_b = F.rotate(img, -270)

    assert_equal(np.array(result_a), np.array(result_b))


@pytest.mark.parametrize("mode", ["L", "RGB", "F"])
def test_rotate_fill(mode):
    img = F.to_pil_image(np.ones((100, 100, 3), dtype=np.uint8) * 255, "RGB")

    num_bands = len(mode)
    wrong_num_bands = num_bands + 1
    fill = 127

    img_conv = img.convert(mode)
    img_rot = F.rotate(img_conv, 45.0, fill=fill)
    pixel = img_rot.getpixel((0, 0))

    if not isinstance(pixel, tuple):
        pixel = (pixel,)
    assert pixel == tuple([fill] * num_bands)

    with pytest.raises(ValueError):
        F.rotate(img_conv, 45.0, fill=tuple([fill] * wrong_num_bands))


def test_gaussian_blur_asserts():
    np_img = np.ones((100, 100, 3), dtype=np.uint8) * 255
    img = F.to_pil_image(np_img, "RGB")

    with pytest.raises(ValueError, match=r"If kernel_size is a sequence its length should be 2"):
        F.gaussian_blur(img, [3])
    with pytest.raises(ValueError, match=r"If kernel_size is a sequence its length should be 2"):
        F.gaussian_blur(img, [3, 3, 3])
    with pytest.raises(ValueError, match=r"Kernel size should be a tuple/list of two integers"):
        transforms.GaussianBlur([3, 3, 3])

    with pytest.raises(ValueError, match=r"kernel_size should have odd and positive integers"):
        F.gaussian_blur(img, [4, 4])
    with pytest.raises(ValueError, match=r"Kernel size value should be an odd and positive number"):
        transforms.GaussianBlur([4, 4])

    with pytest.raises(ValueError, match=r"kernel_size should have odd and positive integers"):
        F.gaussian_blur(img, [-3, -3])
    with pytest.raises(ValueError, match=r"Kernel size value should be an odd and positive number"):
        transforms.GaussianBlur([-3, -3])

    with pytest.raises(ValueError, match=r"If sigma is a sequence, its length should be 2"):
        F.gaussian_blur(img, 3, [1, 1, 1])
    with pytest.raises(ValueError, match=r"sigma should be a single number or a list/tuple with length 2"):
        transforms.GaussianBlur(3, [1, 1, 1])

    with pytest.raises(ValueError, match=r"sigma should have positive values"):
        F.gaussian_blur(img, 3, -1.0)
    with pytest.raises(ValueError, match=r"If sigma is a single number, it must be positive"):
        transforms.GaussianBlur(3, -1.0)

    with pytest.raises(TypeError, match=r"kernel_size should be int or a sequence of integers"):
        F.gaussian_blur(img, "kernel_size_string")
    with pytest.raises(ValueError, match=r"Kernel size should be a tuple/list of two integers"):
        transforms.GaussianBlur("kernel_size_string")

    with pytest.raises(TypeError, match=r"sigma should be either float or sequence of floats"):
        F.gaussian_blur(img, 3, "sigma_string")
    with pytest.raises(ValueError, match=r"sigma should be a single number or a list/tuple with length 2"):
        transforms.GaussianBlur(3, "sigma_string")


def test_lambda():
    trans = transforms.Lambda(lambda x: x.add(10))
    x = torch.randn(10)
    y = trans(x)
    assert_equal(y, torch.add(x, 10))

    trans = transforms.Lambda(lambda x: x.add_(10))
    x = torch.randn(10)
    y = trans(x)
    assert_equal(y, x)

    # Checking if Lambda can be printed as string
    trans.__repr__()


def test_to_grayscale():
    """Unit tests for grayscale transform"""

    x_shape = [2, 2, 3]
    x_data = [0, 5, 13, 54, 135, 226, 37, 8, 234, 90, 255, 1]
    x_np = np.array(x_data, dtype=np.uint8).reshape(x_shape)
    x_pil = Image.fromarray(x_np, mode="RGB")
    x_pil_2 = x_pil.convert("L")
    gray_np = np.array(x_pil_2)

    # Test Set: Grayscale an image with desired number of output channels
    # Case 1: RGB -> 1 channel grayscale
    trans1 = transforms.Grayscale(num_output_channels=1)
    gray_pil_1 = trans1(x_pil)
    gray_np_1 = np.array(gray_pil_1)
    assert gray_pil_1.mode == "L", "mode should be L"
    assert gray_np_1.shape == tuple(x_shape[0:2]), "should be 1 channel"
    assert_equal(gray_np, gray_np_1)

    # Case 2: RGB -> 3 channel grayscale
    trans2 = transforms.Grayscale(num_output_channels=3)
    gray_pil_2 = trans2(x_pil)
    gray_np_2 = np.array(gray_pil_2)
    assert gray_pil_2.mode == "RGB", "mode should be RGB"
    assert gray_np_2.shape == tuple(x_shape), "should be 3 channel"
    assert_equal(gray_np_2[:, :, 0], gray_np_2[:, :, 1])
    assert_equal(gray_np_2[:, :, 1], gray_np_2[:, :, 2])
    assert_equal(gray_np, gray_np_2[:, :, 0])

    # Case 3: 1 channel grayscale -> 1 channel grayscale
    trans3 = transforms.Grayscale(num_output_channels=1)
    gray_pil_3 = trans3(x_pil_2)
    gray_np_3 = np.array(gray_pil_3)
    assert gray_pil_3.mode == "L", "mode should be L"
    assert gray_np_3.shape == tuple(x_shape[0:2]), "should be 1 channel"
    assert_equal(gray_np, gray_np_3)

    # Case 4: 1 channel grayscale -> 3 channel grayscale
    trans4 = transforms.Grayscale(num_output_channels=3)
    gray_pil_4 = trans4(x_pil_2)
    gray_np_4 = np.array(gray_pil_4)
    assert gray_pil_4.mode == "RGB", "mode should be RGB"
    assert gray_np_4.shape == tuple(x_shape), "should be 3 channel"
    assert_equal(gray_np_4[:, :, 0], gray_np_4[:, :, 1])
    assert_equal(gray_np_4[:, :, 1], gray_np_4[:, :, 2])
    assert_equal(gray_np, gray_np_4[:, :, 0])

    # Checking if Grayscale can be printed as string
    trans4.__repr__()


@pytest.mark.parametrize("seed", range(10))
@pytest.mark.parametrize("p", (0, 1))
def test_random_apply(p, seed):
    torch.manual_seed(seed)
    random_apply_transform = transforms.RandomApply([transforms.RandomRotation((45, 50))], p=p)
    img = transforms.ToPILImage()(torch.rand(3, 30, 40))
    out = random_apply_transform(img)
    if p == 0:
        assert out == img
    elif p == 1:
        assert out != img

    # Checking if RandomApply can be printed as string
    random_apply_transform.__repr__()


@pytest.mark.parametrize("seed", range(10))
@pytest.mark.parametrize("proba_passthrough", (0, 1))
def test_random_choice(proba_passthrough, seed):
    random.seed(seed)  # RandomChoice relies on python builtin random.choice, not pytorch

    random_choice_transform = transforms.RandomChoice(
        [
            lambda x: x,  # passthrough
            transforms.RandomRotation((45, 50)),
        ],
        p=[proba_passthrough, 1 - proba_passthrough],
    )

    img = transforms.ToPILImage()(torch.rand(3, 30, 40))
    out = random_choice_transform(img)
    if proba_passthrough == 1:
        assert out == img
    elif proba_passthrough == 0:
        assert out != img

    # Checking if RandomChoice can be printed as string
    random_choice_transform.__repr__()


@pytest.mark.skipif(stats is None, reason="scipy.stats not available")
def test_random_order():
    random_state = random.getstate()
    random.seed(42)
    random_order_transform = transforms.RandomOrder([transforms.Resize(20, antialias=True), transforms.CenterCrop(10)])
    img = transforms.ToPILImage()(torch.rand(3, 25, 25))
    num_samples = 250
    num_normal_order = 0
    resize_crop_out = transforms.CenterCrop(10)(transforms.Resize(20, antialias=True)(img))
    for _ in range(num_samples):
        out = random_order_transform(img)
        if out == resize_crop_out:
            num_normal_order += 1

    p_value = stats.binomtest(num_normal_order, num_samples, p=0.5).pvalue
    random.setstate(random_state)
    assert p_value > 0.0001

    # Checking if RandomOrder can be printed as string
    random_order_transform.__repr__()


def test_linear_transformation():
    num_samples = 1000
    x = torch.randn(num_samples, 3, 10, 10)
    flat_x = x.view(x.size(0), x.size(1) * x.size(2) * x.size(3))
    # compute principal components
    sigma = torch.mm(flat_x.t(), flat_x) / flat_x.size(0)
    u, s, _ = np.linalg.svd(sigma.numpy())
    zca_epsilon = 1e-10  # avoid division by 0
    d = torch.Tensor(np.diag(1.0 / np.sqrt(s + zca_epsilon)))
    u = torch.Tensor(u)
    principal_components = torch.mm(torch.mm(u, d), u.t())
    mean_vector = torch.sum(flat_x, dim=0) / flat_x.size(0)
    # initialize whitening matrix
    whitening = transforms.LinearTransformation(principal_components, mean_vector)
    # estimate covariance and mean using weak law of large number
    num_features = flat_x.size(1)
    cov = 0.0
    mean = 0.0
    for i in x:
        xwhite = whitening(i)
        xwhite = xwhite.view(1, -1).numpy()
        cov += np.dot(xwhite, xwhite.T) / num_features
        mean += np.sum(xwhite) / num_features
    # if rtol for std = 1e-3 then rtol for cov = 2e-3 as std**2 = cov
    torch.testing.assert_close(
        cov / num_samples, np.identity(1), rtol=2e-3, atol=1e-8, check_dtype=False, msg="cov not close to 1"
    )
    torch.testing.assert_close(
        mean / num_samples, 0, rtol=1e-3, atol=1e-8, check_dtype=False, msg="mean not close to 0"
    )

    # Checking if LinearTransformation can be printed as string
    whitening.__repr__()


@pytest.mark.parametrize("dtype", int_dtypes())
def test_max_value(dtype):

    assert F_t._max_value(dtype) == torch.iinfo(dtype).max
    # remove float testing as it can lead to errors such as
    # runtime error: 5.7896e+76 is outside the range of representable values of type 'float'
    # for dtype in float_dtypes():
    # self.assertGreater(F_t._max_value(dtype), torch.finfo(dtype).max)


@pytest.mark.xfail(
    reason="torch.iinfo() is not supported by torchscript. See https://github.com/pytorch/pytorch/issues/41492."
)
def test_max_value_iinfo():
    @torch.jit.script
    def max_value(image: torch.Tensor) -> int:
        return 1 if image.is_floating_point() else torch.iinfo(image.dtype).max


@pytest.mark.parametrize("should_vflip", [True, False])
@pytest.mark.parametrize("single_dim", [True, False])
def test_ten_crop(should_vflip, single_dim):
    to_pil_image = transforms.ToPILImage()
    h = random.randint(5, 25)
    w = random.randint(5, 25)
    crop_h = random.randint(1, h)
    crop_w = random.randint(1, w)
    if single_dim:
        crop_h = min(crop_h, crop_w)
        crop_w = crop_h
        transform = transforms.TenCrop(crop_h, vertical_flip=should_vflip)
        five_crop = transforms.FiveCrop(crop_h)
    else:
        transform = transforms.TenCrop((crop_h, crop_w), vertical_flip=should_vflip)
        five_crop = transforms.FiveCrop((crop_h, crop_w))

    img = to_pil_image(torch.FloatTensor(3, h, w).uniform_())
    results = transform(img)
    expected_output = five_crop(img)

    # Checking if FiveCrop and TenCrop can be printed as string
    transform.__repr__()
    five_crop.__repr__()

    if should_vflip:
        vflipped_img = img.transpose(Image.FLIP_TOP_BOTTOM)
        expected_output += five_crop(vflipped_img)
    else:
        hflipped_img = img.transpose(Image.FLIP_LEFT_RIGHT)
        expected_output += five_crop(hflipped_img)

    assert len(results) == 10
    assert results == expected_output


@pytest.mark.parametrize("single_dim", [True, False])
def test_five_crop(single_dim):
    to_pil_image = transforms.ToPILImage()
    h = random.randint(5, 25)
    w = random.randint(5, 25)
    crop_h = random.randint(1, h)
    crop_w = random.randint(1, w)
    if single_dim:
        crop_h = min(crop_h, crop_w)
        crop_w = crop_h
        transform = transforms.FiveCrop(crop_h)
    else:
        transform = transforms.FiveCrop((crop_h, crop_w))

    img = torch.FloatTensor(3, h, w).uniform_()

    results = transform(to_pil_image(img))

    assert len(results) == 5
    for crop in results:
        assert crop.size == (crop_w, crop_h)

    to_pil_image = transforms.ToPILImage()
    tl = to_pil_image(img[:, 0:crop_h, 0:crop_w])
    tr = to_pil_image(img[:, 0:crop_h, w - crop_w :])
    bl = to_pil_image(img[:, h - crop_h :, 0:crop_w])
    br = to_pil_image(img[:, h - crop_h :, w - crop_w :])
    center = transforms.CenterCrop((crop_h, crop_w))(to_pil_image(img))
    expected_output = (tl, tr, bl, br, center)
    assert results == expected_output


@pytest.mark.parametrize("policy", transforms.AutoAugmentPolicy)
@pytest.mark.parametrize("fill", [None, 85, (128, 128, 128)])
@pytest.mark.parametrize("grayscale", [True, False])
def test_autoaugment(policy, fill, grayscale):
    random.seed(42)
    img = Image.open(GRACE_HOPPER)
    if grayscale:
        img, fill = _get_grayscale_test_image(img, fill)
    transform = transforms.AutoAugment(policy=policy, fill=fill)
    for _ in range(100):
        img = transform(img)
    transform.__repr__()


@pytest.mark.parametrize("num_ops", [1, 2, 3])
@pytest.mark.parametrize("magnitude", [7, 9, 11])
@pytest.mark.parametrize("fill", [None, 85, (128, 128, 128)])
@pytest.mark.parametrize("grayscale", [True, False])
def test_randaugment(num_ops, magnitude, fill, grayscale):
    random.seed(42)
    img = Image.open(GRACE_HOPPER)
    if grayscale:
        img, fill = _get_grayscale_test_image(img, fill)
    transform = transforms.RandAugment(num_ops=num_ops, magnitude=magnitude, fill=fill)
    for _ in range(100):
        img = transform(img)
    transform.__repr__()


@pytest.mark.parametrize("fill", [None, 85, (128, 128, 128)])
@pytest.mark.parametrize("num_magnitude_bins", [10, 13, 30])
@pytest.mark.parametrize("grayscale", [True, False])
def test_trivialaugmentwide(fill, num_magnitude_bins, grayscale):
    random.seed(42)
    img = Image.open(GRACE_HOPPER)
    if grayscale:
        img, fill = _get_grayscale_test_image(img, fill)
    transform = transforms.TrivialAugmentWide(fill=fill, num_magnitude_bins=num_magnitude_bins)
    for _ in range(100):
        img = transform(img)
    transform.__repr__()


@pytest.mark.parametrize("fill", [None, 85, (128, 128, 128)])
@pytest.mark.parametrize("severity", [1, 10])
@pytest.mark.parametrize("mixture_width", [1, 2])
@pytest.mark.parametrize("chain_depth", [-1, 2])
@pytest.mark.parametrize("all_ops", [True, False])
@pytest.mark.parametrize("grayscale", [True, False])
def test_augmix(fill, severity, mixture_width, chain_depth, all_ops, grayscale):
    random.seed(42)
    img = Image.open(GRACE_HOPPER)
    if grayscale:
        img, fill = _get_grayscale_test_image(img, fill)
    transform = transforms.AugMix(
        fill=fill, severity=severity, mixture_width=mixture_width, chain_depth=chain_depth, all_ops=all_ops
    )
    for _ in range(100):
        img = transform(img)
    transform.__repr__()


def test_random_crop():
    height = random.randint(10, 32) * 2
    width = random.randint(10, 32) * 2
    oheight = random.randint(5, (height - 2) // 2) * 2
    owidth = random.randint(5, (width - 2) // 2) * 2
    img = torch.ones(3, height, width, dtype=torch.uint8)
    result = transforms.Compose(
        [
            transforms.ToPILImage(),
            transforms.RandomCrop((oheight, owidth)),
            transforms.PILToTensor(),
        ]
    )(img)
    assert result.size(1) == oheight
    assert result.size(2) == owidth

    padding = random.randint(1, 20)
    result = transforms.Compose(
        [
            transforms.ToPILImage(),
            transforms.RandomCrop((oheight, owidth), padding=padding),
            transforms.PILToTensor(),
        ]
    )(img)
    assert result.size(1) == oheight
    assert result.size(2) == owidth

    result = transforms.Compose(
        [transforms.ToPILImage(), transforms.RandomCrop((height, width)), transforms.PILToTensor()]
    )(img)
    assert result.size(1) == height
    assert result.size(2) == width
    torch.testing.assert_close(result, img)

    result = transforms.Compose(
        [
            transforms.ToPILImage(),
            transforms.RandomCrop((height + 1, width + 1), pad_if_needed=True),
            transforms.PILToTensor(),
        ]
    )(img)
    assert result.size(1) == height + 1
    assert result.size(2) == width + 1

    t = transforms.RandomCrop(33)
    img = torch.ones(3, 32, 32)
    with pytest.raises(ValueError, match=r"Required crop size .+ is larger than input image size .+"):
        t(img)


def test_center_crop():
    height = random.randint(10, 32) * 2
    width = random.randint(10, 32) * 2
    oheight = random.randint(5, (height - 2) // 2) * 2
    owidth = random.randint(5, (width - 2) // 2) * 2

    img = torch.ones(3, height, width, dtype=torch.uint8)
    oh1 = (height - oheight) // 2
    ow1 = (width - owidth) // 2
    imgnarrow = img[:, oh1 : oh1 + oheight, ow1 : ow1 + owidth]
    imgnarrow.fill_(0)
    result = transforms.Compose(
        [
            transforms.ToPILImage(),
            transforms.CenterCrop((oheight, owidth)),
            transforms.PILToTensor(),
        ]
    )(img)
    assert result.sum() == 0
    oheight += 1
    owidth += 1
    result = transforms.Compose(
        [
            transforms.ToPILImage(),
            transforms.CenterCrop((oheight, owidth)),
            transforms.PILToTensor(),
        ]
    )(img)
    sum1 = result.sum()
    assert sum1 > 1
    oheight += 1
    owidth += 1
    result = transforms.Compose(
        [
            transforms.ToPILImage(),
            transforms.CenterCrop((oheight, owidth)),
            transforms.PILToTensor(),
        ]
    )(img)
    sum2 = result.sum()
    assert sum2 > 0
    assert sum2 > sum1


@pytest.mark.parametrize("odd_image_size", (True, False))
@pytest.mark.parametrize("delta", (1, 3, 5))
@pytest.mark.parametrize("delta_width", (-2, -1, 0, 1, 2))
@pytest.mark.parametrize("delta_height", (-2, -1, 0, 1, 2))
def test_center_crop_2(odd_image_size, delta, delta_width, delta_height):
    """Tests when center crop size is larger than image size, along any dimension"""

    # Since height is independent of width, we can ignore images with odd height and even width and vice-versa.
    input_image_size = (random.randint(10, 32) * 2, random.randint(10, 32) * 2)
    if odd_image_size:
        input_image_size = (input_image_size[0] + 1, input_image_size[1] + 1)

    delta_height *= delta
    delta_width *= delta

    img = torch.ones(3, *input_image_size, dtype=torch.uint8)
    crop_size = (input_image_size[0] + delta_height, input_image_size[1] + delta_width)

    # Test both transforms, one with PIL input and one with tensor
    output_pil = transforms.Compose(
        [transforms.ToPILImage(), transforms.CenterCrop(crop_size), transforms.PILToTensor()],
    )(img)
    assert output_pil.size()[1:3] == crop_size

    output_tensor = transforms.CenterCrop(crop_size)(img)
    assert output_tensor.size()[1:3] == crop_size

    # Ensure output for PIL and Tensor are equal
    assert_equal(
        output_tensor,
        output_pil,
        msg=f"image_size: {input_image_size} crop_size: {crop_size}",
    )

    # Check if content in center of both image and cropped output is same.
    center_size = (min(crop_size[0], input_image_size[0]), min(crop_size[1], input_image_size[1]))
    crop_center_tl, input_center_tl = [0, 0], [0, 0]
    for index in range(2):
        if crop_size[index] > input_image_size[index]:
            crop_center_tl[index] = (crop_size[index] - input_image_size[index]) // 2
        else:
            input_center_tl[index] = (input_image_size[index] - crop_size[index]) // 2

    output_center = output_pil[
        :,
        crop_center_tl[0] : crop_center_tl[0] + center_size[0],
        crop_center_tl[1] : crop_center_tl[1] + center_size[1],
    ]

    img_center = img[
        :,
        input_center_tl[0] : input_center_tl[0] + center_size[0],
        input_center_tl[1] : input_center_tl[1] + center_size[1],
    ]

    assert_equal(output_center, img_center)


def test_color_jitter():
    color_jitter = transforms.ColorJitter(2, 2, 2, 0.1)

    x_shape = [2, 2, 3]
    x_data = [0, 5, 13, 54, 135, 226, 37, 8, 234, 90, 255, 1]
    x_np = np.array(x_data, dtype=np.uint8).reshape(x_shape)
    x_pil = Image.fromarray(x_np, mode="RGB")
    x_pil_2 = x_pil.convert("L")

    for _ in range(10):
        y_pil = color_jitter(x_pil)
        assert y_pil.mode == x_pil.mode

        y_pil_2 = color_jitter(x_pil_2)
        assert y_pil_2.mode == x_pil_2.mode

    # Checking if ColorJitter can be printed as string
    color_jitter.__repr__()


@pytest.mark.parametrize("hue", [1, (-1, 1)])
def test_color_jitter_hue_out_of_bounds(hue):
    with pytest.raises(ValueError, match=re.escape("hue values should be between (-0.5, 0.5)")):
        transforms.ColorJitter(hue=hue)


@pytest.mark.parametrize("seed", range(10))
@pytest.mark.skipif(stats is None, reason="scipy.stats not available")
def test_random_erasing(seed):
    torch.random.manual_seed(seed)
    img = torch.ones(3, 128, 128)

    t = transforms.RandomErasing(scale=(0.1, 0.1), ratio=(1 / 3, 3.0))
    y, x, h, w, v = t.get_params(
        img,
        t.scale,
        t.ratio,
        [
            t.value,
        ],
    )
    aspect_ratio = h / w
    # Add some tolerance due to the rounding and int conversion used in the transform
    tol = 0.05
    assert 1 / 3 - tol <= aspect_ratio <= 3 + tol

    # Make sure that h > w and h < w are equally likely (log-scale sampling)
    aspect_ratios = []
    random.seed(42)
    trial = 1000
    for _ in range(trial):
        y, x, h, w, v = t.get_params(
            img,
            t.scale,
            t.ratio,
            [
                t.value,
            ],
        )
        aspect_ratios.append(h / w)

    count_bigger_then_ones = len([1 for aspect_ratio in aspect_ratios if aspect_ratio > 1])
    p_value = stats.binomtest(count_bigger_then_ones, trial, p=0.5).pvalue
    assert p_value > 0.0001

    # Checking if RandomErasing can be printed as string
    t.__repr__()


def test_random_rotation():

    with pytest.raises(ValueError):
        transforms.RandomRotation(-0.7)

    with pytest.raises(ValueError):
        transforms.RandomRotation([-0.7])

    with pytest.raises(ValueError):
        transforms.RandomRotation([-0.7, 0, 0.7])

    t = transforms.RandomRotation(0, fill=None)
    assert t.fill == 0

    t = transforms.RandomRotation(10)
    angle = t.get_params(t.degrees)
    assert angle > -10 and angle < 10

    t = transforms.RandomRotation((-10, 10))
    angle = t.get_params(t.degrees)
    assert -10 < angle < 10

    # Checking if RandomRotation can be printed as string
    t.__repr__()

    t = transforms.RandomRotation((-10, 10), interpolation=Image.BILINEAR)
    assert t.interpolation == transforms.InterpolationMode.BILINEAR


def test_random_rotation_error():
    # assert fill being either a Sequence or a Number
    with pytest.raises(TypeError):
        transforms.RandomRotation(0, fill={})


def test_randomperspective():
    for _ in range(10):
        height = random.randint(24, 32) * 2
        width = random.randint(24, 32) * 2
        img = torch.ones(3, height, width)
        to_pil_image = transforms.ToPILImage()
        img = to_pil_image(img)
        perp = transforms.RandomPerspective()
        startpoints, endpoints = perp.get_params(width, height, 0.5)
        tr_img = F.perspective(img, startpoints, endpoints)
        tr_img2 = F.convert_image_dtype(F.pil_to_tensor(F.perspective(tr_img, endpoints, startpoints)))
        tr_img = F.convert_image_dtype(F.pil_to_tensor(tr_img))
        assert img.size[0] == width
        assert img.size[1] == height
        assert torch.nn.functional.mse_loss(
            tr_img, F.convert_image_dtype(F.pil_to_tensor(img))
        ) + 0.3 > torch.nn.functional.mse_loss(tr_img2, F.convert_image_dtype(F.pil_to_tensor(img)))


@pytest.mark.parametrize("seed", range(10))
@pytest.mark.parametrize("mode", ["L", "RGB", "F"])
def test_randomperspective_fill(mode, seed):
    torch.random.manual_seed(seed)

    # assert fill being either a Sequence or a Number
    with pytest.raises(TypeError):
        transforms.RandomPerspective(fill={})

    t = transforms.RandomPerspective(fill=None)
    assert t.fill == 0

    height = 100
    width = 100
    img = torch.ones(3, height, width)
    to_pil_image = transforms.ToPILImage()
    img = to_pil_image(img)
    fill = 127
    num_bands = len(mode)

    img_conv = img.convert(mode)
    perspective = transforms.RandomPerspective(p=1, fill=fill)
    tr_img = perspective(img_conv)
    pixel = tr_img.getpixel((0, 0))

    if not isinstance(pixel, tuple):
        pixel = (pixel,)
    assert pixel == tuple([fill] * num_bands)

    startpoints, endpoints = transforms.RandomPerspective.get_params(width, height, 0.5)
    tr_img = F.perspective(img_conv, startpoints, endpoints, fill=fill)
    pixel = tr_img.getpixel((0, 0))

    if not isinstance(pixel, tuple):
        pixel = (pixel,)
    assert pixel == tuple([fill] * num_bands)

    wrong_num_bands = num_bands + 1
    with pytest.raises(ValueError):
        F.perspective(img_conv, startpoints, endpoints, fill=tuple([fill] * wrong_num_bands))


@pytest.mark.skipif(stats is None, reason="scipy.stats not available")
def test_normalize():
    def samples_from_standard_normal(tensor):
        p_value = stats.kstest(list(tensor.view(-1)), "norm", args=(0, 1)).pvalue
        return p_value > 0.0001

    random_state = random.getstate()
    random.seed(42)
    for channels in [1, 3]:
        img = torch.rand(channels, 10, 10)
        mean = [img[c].mean() for c in range(channels)]
        std = [img[c].std() for c in range(channels)]
        normalized = transforms.Normalize(mean, std)(img)
        assert samples_from_standard_normal(normalized)
    random.setstate(random_state)

    # Checking if Normalize can be printed as string
    transforms.Normalize(mean, std).__repr__()

    # Checking the optional in-place behaviour
    tensor = torch.rand((1, 16, 16))
    tensor_inplace = transforms.Normalize((0.5,), (0.5,), inplace=True)(tensor)
    assert_equal(tensor, tensor_inplace)


@pytest.mark.parametrize("dtype1", [torch.float32, torch.float64])
@pytest.mark.parametrize("dtype2", [torch.int64, torch.float32, torch.float64])
def test_normalize_different_dtype(dtype1, dtype2):
    img = torch.rand(3, 10, 10, dtype=dtype1)
    mean = torch.tensor([1, 2, 3], dtype=dtype2)
    std = torch.tensor([1, 2, 1], dtype=dtype2)
    # checks that it doesn't crash
    transforms.functional.normalize(img, mean, std)


def test_normalize_3d_tensor():
    torch.manual_seed(28)
    n_channels = 3
    img_size = 10
    mean = torch.rand(n_channels)
    std = torch.rand(n_channels)
    img = torch.rand(n_channels, img_size, img_size)
    target = F.normalize(img, mean, std)

    mean_unsqueezed = mean.view(-1, 1, 1)
    std_unsqueezed = std.view(-1, 1, 1)
    result1 = F.normalize(img, mean_unsqueezed, std_unsqueezed)
    result2 = F.normalize(
        img, mean_unsqueezed.repeat(1, img_size, img_size), std_unsqueezed.repeat(1, img_size, img_size)
    )
    torch.testing.assert_close(target, result1)
    torch.testing.assert_close(target, result2)


class TestAffine:
    @pytest.fixture(scope="class")
    def input_img(self):
        input_img = np.zeros((40, 40, 3), dtype=np.uint8)
        for pt in [(16, 16), (20, 16), (20, 20)]:
            for i in range(-5, 5):
                for j in range(-5, 5):
                    input_img[pt[0] + i, pt[1] + j, :] = [255, 155, 55]
        return input_img

    def test_affine_translate_seq(self, input_img):
        with pytest.raises(TypeError, match=r"Argument translate should be a sequence"):
            F.affine(input_img, 10, translate=0, scale=1, shear=1)

    @pytest.fixture(scope="class")
    def pil_image(self, input_img):
        return F.to_pil_image(input_img)

    def _to_3x3_inv(self, inv_result_matrix):
        result_matrix = np.zeros((3, 3))
        result_matrix[:2, :] = np.array(inv_result_matrix).reshape((2, 3))
        result_matrix[2, 2] = 1
        return np.linalg.inv(result_matrix)

    def _test_transformation(self, angle, translate, scale, shear, pil_image, input_img, center=None):

        a_rad = math.radians(angle)
        s_rad = [math.radians(sh_) for sh_ in shear]
        cnt = [20, 20] if center is None else center
        cx, cy = cnt
        tx, ty = translate
        sx, sy = s_rad
        rot = a_rad

        # 1) Check transformation matrix:
        C = np.array([[1, 0, cx], [0, 1, cy], [0, 0, 1]])
        T = np.array([[1, 0, tx], [0, 1, ty], [0, 0, 1]])
        Cinv = np.linalg.inv(C)

        RS = np.array(
            [
                [scale * math.cos(rot), -scale * math.sin(rot), 0],
                [scale * math.sin(rot), scale * math.cos(rot), 0],
                [0, 0, 1],
            ]
        )

        SHx = np.array([[1, -math.tan(sx), 0], [0, 1, 0], [0, 0, 1]])

        SHy = np.array([[1, 0, 0], [-math.tan(sy), 1, 0], [0, 0, 1]])

        RSS = np.matmul(RS, np.matmul(SHy, SHx))

        true_matrix = np.matmul(T, np.matmul(C, np.matmul(RSS, Cinv)))

        result_matrix = self._to_3x3_inv(
            F._get_inverse_affine_matrix(center=cnt, angle=angle, translate=translate, scale=scale, shear=shear)
        )
        assert np.sum(np.abs(true_matrix - result_matrix)) < 1e-10
        # 2) Perform inverse mapping:
        true_result = np.zeros((40, 40, 3), dtype=np.uint8)
        inv_true_matrix = np.linalg.inv(true_matrix)
        for y in range(true_result.shape[0]):
            for x in range(true_result.shape[1]):
                # Same as for PIL:
                # https://github.com/python-pillow/Pillow/blob/71f8ec6a0cfc1008076a023c0756542539d057ab/
                # src/libImaging/Geometry.c#L1060
                input_pt = np.array([x + 0.5, y + 0.5, 1.0])
                res = np.floor(np.dot(inv_true_matrix, input_pt)).astype(int)
                _x, _y = res[:2]
                if 0 <= _x < input_img.shape[1] and 0 <= _y < input_img.shape[0]:
                    true_result[y, x, :] = input_img[_y, _x, :]

        result = F.affine(pil_image, angle=angle, translate=translate, scale=scale, shear=shear, center=center)
        assert result.size == pil_image.size
        # Compute number of different pixels:
        np_result = np.array(result)
        n_diff_pixels = np.sum(np_result != true_result) / 3
        # Accept 3 wrong pixels
        error_msg = (
            f"angle={angle}, translate={translate}, scale={scale}, shear={shear}\nn diff pixels={n_diff_pixels}\n"
        )
        assert n_diff_pixels < 3, error_msg

    def test_transformation_discrete(self, pil_image, input_img):
        # Test rotation
        angle = 45
        self._test_transformation(
            angle=angle, translate=(0, 0), scale=1.0, shear=(0.0, 0.0), pil_image=pil_image, input_img=input_img
        )

        # Test rotation
        angle = 45
        self._test_transformation(
            angle=angle,
            translate=(0, 0),
            scale=1.0,
            shear=(0.0, 0.0),
            pil_image=pil_image,
            input_img=input_img,
            center=[0, 0],
        )

        # Test translation
        translate = [10, 15]
        self._test_transformation(
            angle=0.0, translate=translate, scale=1.0, shear=(0.0, 0.0), pil_image=pil_image, input_img=input_img
        )

        # Test scale
        scale = 1.2
        self._test_transformation(
            angle=0.0, translate=(0.0, 0.0), scale=scale, shear=(0.0, 0.0), pil_image=pil_image, input_img=input_img
        )

        # Test shear
        shear = [45.0, 25.0]
        self._test_transformation(
            angle=0.0, translate=(0.0, 0.0), scale=1.0, shear=shear, pil_image=pil_image, input_img=input_img
        )

        # Test shear with top-left as center
        shear = [45.0, 25.0]
        self._test_transformation(
            angle=0.0,
            translate=(0.0, 0.0),
            scale=1.0,
            shear=shear,
            pil_image=pil_image,
            input_img=input_img,
            center=[0, 0],
        )

    @pytest.mark.parametrize("angle", range(-90, 90, 36))
    @pytest.mark.parametrize("translate", range(-10, 10, 5))
    @pytest.mark.parametrize("scale", [0.77, 1.0, 1.27])
    @pytest.mark.parametrize("shear", range(-15, 15, 5))
    def test_transformation_range(self, angle, translate, scale, shear, pil_image, input_img):
        self._test_transformation(
            angle=angle,
            translate=(translate, translate),
            scale=scale,
            shear=(shear, shear),
            pil_image=pil_image,
            input_img=input_img,
        )


def test_random_affine():

    with pytest.raises(ValueError):
        transforms.RandomAffine(-0.7)
    with pytest.raises(ValueError):
        transforms.RandomAffine([-0.7])
    with pytest.raises(ValueError):
        transforms.RandomAffine([-0.7, 0, 0.7])
    with pytest.raises(TypeError):
        transforms.RandomAffine([-90, 90], translate=2.0)
    with pytest.raises(ValueError):
        transforms.RandomAffine([-90, 90], translate=[-1.0, 1.0])
    with pytest.raises(ValueError):
        transforms.RandomAffine([-90, 90], translate=[-1.0, 0.0, 1.0])

    with pytest.raises(ValueError):
        transforms.RandomAffine([-90, 90], translate=[0.2, 0.2], scale=[0.0])
    with pytest.raises(ValueError):
        transforms.RandomAffine([-90, 90], translate=[0.2, 0.2], scale=[-1.0, 1.0])
    with pytest.raises(ValueError):
        transforms.RandomAffine([-90, 90], translate=[0.2, 0.2], scale=[0.5, -0.5])
    with pytest.raises(ValueError):
        transforms.RandomAffine([-90, 90], translate=[0.2, 0.2], scale=[0.5, 3.0, -0.5])

    with pytest.raises(ValueError):
        transforms.RandomAffine([-90, 90], translate=[0.2, 0.2], scale=[0.5, 0.5], shear=-7)
    with pytest.raises(ValueError):
        transforms.RandomAffine([-90, 90], translate=[0.2, 0.2], scale=[0.5, 0.5], shear=[-10])
    with pytest.raises(ValueError):
        transforms.RandomAffine([-90, 90], translate=[0.2, 0.2], scale=[0.5, 0.5], shear=[-10, 0, 10])
    with pytest.raises(ValueError):
        transforms.RandomAffine([-90, 90], translate=[0.2, 0.2], scale=[0.5, 0.5], shear=[-10, 0, 10, 0, 10])

    # assert fill being either a Sequence or a Number
    with pytest.raises(TypeError):
        transforms.RandomAffine(0, fill={})

    t = transforms.RandomAffine(0, fill=None)
    assert t.fill == 0

    x = np.zeros((100, 100, 3), dtype=np.uint8)
    img = F.to_pil_image(x)

    t = transforms.RandomAffine(10, translate=[0.5, 0.3], scale=[0.7, 1.3], shear=[-10, 10, 20, 40])
    for _ in range(100):
        angle, translations, scale, shear = t.get_params(t.degrees, t.translate, t.scale, t.shear, img_size=img.size)
        assert -10 < angle < 10
        assert -img.size[0] * 0.5 <= translations[0] <= img.size[0] * 0.5
        assert -img.size[1] * 0.5 <= translations[1] <= img.size[1] * 0.5
        assert 0.7 < scale < 1.3
        assert -10 < shear[0] < 10
        assert -20 < shear[1] < 40

    # Checking if RandomAffine can be printed as string
    t.__repr__()

    t = transforms.RandomAffine(10, interpolation=transforms.InterpolationMode.BILINEAR)
    assert "bilinear" in t.__repr__()

    t = transforms.RandomAffine(10, interpolation=Image.BILINEAR)
    assert t.interpolation == transforms.InterpolationMode.BILINEAR


def test_elastic_transformation():
    with pytest.raises(TypeError, match=r"alpha should be float or a sequence of floats"):
        transforms.ElasticTransform(alpha=True, sigma=2.0)
    with pytest.raises(TypeError, match=r"alpha should be a sequence of floats"):
        transforms.ElasticTransform(alpha=[1.0, True], sigma=2.0)
    with pytest.raises(ValueError, match=r"alpha is a sequence its length should be 2"):
        transforms.ElasticTransform(alpha=[1.0, 0.0, 1.0], sigma=2.0)

    with pytest.raises(TypeError, match=r"sigma should be float or a sequence of floats"):
        transforms.ElasticTransform(alpha=2.0, sigma=True)
    with pytest.raises(TypeError, match=r"sigma should be a sequence of floats"):
        transforms.ElasticTransform(alpha=2.0, sigma=[1.0, True])
    with pytest.raises(ValueError, match=r"sigma is a sequence its length should be 2"):
        transforms.ElasticTransform(alpha=2.0, sigma=[1.0, 0.0, 1.0])

    t = transforms.transforms.ElasticTransform(alpha=2.0, sigma=2.0, interpolation=Image.BILINEAR)
    assert t.interpolation == transforms.InterpolationMode.BILINEAR

    with pytest.raises(TypeError, match=r"fill should be int or float"):
        transforms.ElasticTransform(alpha=1.0, sigma=1.0, fill={})

    x = torch.randint(0, 256, (3, 32, 32), dtype=torch.uint8)
    img = F.to_pil_image(x)
    t = transforms.ElasticTransform(alpha=0.0, sigma=0.0)
    transformed_img = t(img)
    assert transformed_img == img

    # Smoke test on PIL images
    t = transforms.ElasticTransform(alpha=0.5, sigma=0.23)
    transformed_img = t(img)
    assert isinstance(transformed_img, Image.Image)

    # Checking if ElasticTransform can be printed as string
    t.__repr__()


def test_random_grayscale_with_grayscale_input():
    transform = transforms.RandomGrayscale(p=1.0)

    image_tensor = torch.randint(0, 256, (1, 16, 16), dtype=torch.uint8)
    output_tensor = transform(image_tensor)
    torch.testing.assert_close(output_tensor, image_tensor)

    image_pil = F.to_pil_image(image_tensor)
    output_pil = transform(image_pil)
    torch.testing.assert_close(F.pil_to_tensor(output_pil), image_tensor)


if __name__ == "__main__":
    pytest.main([__file__])



from typing import Any, Dict

import torch
from torchvision.transforms.v2 import functional as F, Transform


class UniformTemporalSubsample(Transform):
    """Uniformly subsample ``num_samples`` indices from the temporal dimension of the video.

    Videos are expected to be of shape ``[..., T, C, H, W]`` where ``T`` denotes the temporal dimension.

    When ``num_samples`` is larger than the size of temporal dimension of the video, it
    will sample frames based on nearest neighbor interpolation.

    Args:
        num_samples (int): The number of equispaced samples to be selected
    """

    _transformed_types = (torch.Tensor,)

    def __init__(self, num_samples: int):
        super().__init__()
        self.num_samples = num_samples

    def _transform(self, inpt: Any, params: Dict[str, Any]) -> Any:
        return self._call_kernel(F.uniform_temporal_subsample, inpt, self.num_samples)
